# -*- coding: utf-8 -*-
"""Models for Python objects in Tanium's API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import json
import operator
import re
import six
import warnings

from . import exceptions

float_types = (float,)
""":obj:`tuple` of :obj:`type`: Float types."""

integer_types = six.integer_types
""":obj:`tuple` of :obj:`type`: Integer types."""

string_types = six.string_types
""":obj:`tuple` of :obj:`type`: String types."""

simple_types = tuple(list(integer_types) + list(string_types) + list(float_types))
""":obj:`tuple` of :obj:`type`: All types that should be considered simple types."""


class ApiModel(object):
    """Base class for all models in the API."""

    API_NAME = None
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = None
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = None
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = None
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = None
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = None
    """:obj:`list` or :obj:`str`: Added to :attr:`API_STR` in str."""

    T_ATTR_REPR = "{attr}={value!r}"
    """:obj:`str`: Template for simple attrs in str and all attr in repr."""

    T_ATTR_STR = "{attr}={value}"
    """:obj:`str`: Template for complex attrs in str."""

    T_ATTR_JOIN = ", "
    """:obj:`str`: String used to join attr vals in str and repr."""

    T_CLS_STR = "{obj.__class__.__name__}({attrs})"
    """:obj:`str`: Template for class name and attrs in str."""

    T_CLS_REPR = "{obj.__class__.__name__}({attrs})"
    """:obj:`str`: Template for class name and attrs in repr."""

    T_ATTR_DESC = "  - {api_type} attribute {attr!r} of type {attr_type!r}"
    """:obj:`str`: Template for attr line in :meth:`api_attrs_desc`."""

    T_DESC = "Defined attributes:"
    """:obj:`str`: Template for first line in :meth:`api_attrs_desc`."""

    API_STR = [
        "id",
        "name",
        "display_name",
        "value",
        "type",
        "public_flag",
        "hidden_flag",
        "question",
        "query_text",
        "question_text",
        "expiration",
        "saved_question",
        "from_canonical_text",
    ]
    """:obj:`list` of :obj:`str`: Attrs to display in str or to put first in repr."""

    def __copy__(self):
        """Support shallow copy for self.

        Returns:
            :obj:`ApiItem`

        """
        return self.__class__(**self.__dict__)

    def __deepcopy__(self, memo):
        """Support deep copy for self.

        Returns:
            :obj:`ApiItem`

        """
        self_dict = self.serialize(wrap_name=False)
        new_dict = copy.deepcopy(self_dict, memo)
        return self.__class__(**new_dict)

    def __eq__(self, value):
        """Support for self == value.

        Args:
            value (:obj:`dict` or :obj:`list` or :obj:`ApiModel` or :obj:`object`):
                Value for comparison.

        Notes:
            * If value is a dict, it will be turned into an ApiModel.
            * If this object is an ApiList and value is a list, it will be turned
              into an ApiList.
            * If value is turned into or already is an ApiModel, it will be serialized
              into a dict/list.
            * Finally, actual comparison of value is done against this object
              serialized into a dict/list.

        Returns:
            :obj:`bool`

        """
        sargs = {"wrap_name": False, "wrap_item_attr": False}
        this = self.serialize(**sargs)

        if isinstance(value, dict):
            value = self.__class__(**value)
        elif isinstance(value, list) and isinstance(self, ApiList):
            value = self.__class__(*value)

        if isinstance(value, ApiModel):
            value = value.serialize(**sargs)

        return this == value

    def __ne__(self, value):
        """Support for self != value.

        Args:
            value (:obj:`dict` or :obj:`list` or :obj:`ApiModel` or :obj:`object`):
                Value for comparison.

        Returns:
            :obj:`bool`

        """
        return not self == value

    @classmethod
    def api_attrs(cls):
        """Get simple and complex attributes combined.

        Returns:
            :obj:`dict`

        """
        return dict(list(cls.API_SIMPLE.items()) + list(cls.API_COMPLEX.items()))

    @classmethod
    def api_attrs_repr(cls):
        """Get attributes for repr formatting.

        Notes:
            Will return attrs without dupes in order of priority to show on
            format string from :attr:`API_STR`, :attr:`API_STR_ADD`,
            and :meth:`api_attrs`.

        Returns:
            :obj:`list` of :obj:`str`

        """
        api_attrs = cls.api_attrs()
        show_attrs = cls.API_STR + cls.API_STR_ADD + list(api_attrs)
        attrs = []
        for attr in show_attrs:
            if attr not in attrs and (attr in api_attrs or hasattr(cls, attr)):
                attrs.append(attr)
        return attrs or list(api_attrs)

    @classmethod
    def api_attrs_str(cls):
        """Get attributes for str formatting.

        Notes:
            Will return attrs without dupes in order of priority to show on
            format string from :attr:`API_STR` and :attr:`API_STR_ADD`.

            If no attrs from :attr:`API_STR` and :attr:`API_STR_ADD` exist
            in :meth:`api_attrs`, return all attrs from :meth:`api_attrs`.

        Returns:
            :obj:`list` of :obj:`str`

        """
        api_attrs = cls.api_attrs()
        show_attrs = cls.API_STR + cls.API_STR_ADD
        attrs = []
        for attr in show_attrs:
            if attr not in attrs and (attr in api_attrs or hasattr(cls, attr)):
                attrs.append(attr)
        return attrs or list(api_attrs)

    @classmethod
    def api_attrs_desc(cls):
        """Get description of attrs from :attr:`API_SIMPLE` and :attr:`API_COMPLEX`.

        Notes:
            Used by exceptions to add defined attributes to error messages.

        Returns:
            :obj:`list` of :obj:`str`

        """
        tmpl = cls.T_ATTR_DESC.format
        lines = []
        lines.append(cls.T_DESC)
        for attr, attr_type in cls.API_SIMPLE.items():
            desc = tmpl(api_type="simple", attr=attr, attr_type=attr_type)
            lines.append(desc)
        for attr, attr_type in cls.API_COMPLEX.items():
            desc = tmpl(api_type="complex", attr=attr, attr_type=attr_type)
            lines.append(desc)
        return lines

    @classmethod
    def api_coerce_int(cls, value):
        """Try to coerce value into :obj:`int` if possible.

        Args:
            value (:obj:`object`):
                Object to coerce into int.

        Returns:
            :obj:`int` or :obj:`object` if unchanged.

        """
        try:
            return int(value)
        except Exception:
            return None if value == "" else value
            # REST GetResultData can return "max_available_age" of ""

    @classmethod
    def api_coerce_float(cls, value):
        """Try to coerce value into :obj:`float` if possible.

        Args:
            value (:obj:`object`):
                Object to coerce into float.

        Returns:
            :obj:`float` or :obj:`object` if unchanged.

        """
        try:
            return float(value)
        except Exception:
            return value

    @classmethod
    def api_coerce_list(cls, value):
        """Coerce value into :obj:`list` if it is not already.

        Args:
            value (:obj:`object`):
                Object to coerce into list. If value is None, will return empty list.

        Returns:
            :obj:`list` or :obj:`object` if unchanged.

        """
        if value is None:
            return []
        if not isinstance(value, (list, tuple)):
            return [value]
        return value

    @classmethod
    def api_coerce_simple(cls, value, be_type):
        """Try to coerce a value into a simple type.

        Args:
            value (:obj:`object`):
                Value that should be coerced.
            be_type (:class:`object`):
                Type that value should be coerced to.

        Returns:
            :obj:`float` or :obj:`int` or :obj:`str` or :obj:`object` if unchanged.

        """
        if be_type == float_types and not isinstance(value, float_types):
            return cls.api_coerce_float(value=value)
        elif be_type == integer_types:
            if isinstance(value, bool) or not isinstance(value, integer_types):
                return cls.api_coerce_int(value=value)
        elif be_type == string_types and isinstance(value, simple_types):
            return format(value)
        return value

    @classmethod
    def api_coerce_complex(cls, value, be_type):
        """Try to coerce a value into a complex type.

        Args:
            value (:obj:`object`):
                Value that should be coerced.
            be_type (:class:`object`):
                Type that value should be coerced to.

        Returns:
            :obj:`ApiItem` or :obj:`ApiList` or :obj:`object` if unchanged.

        """
        if not isinstance(value, be_type):
            if isinstance(value, dict):
                return be_type(**value)
            elif isinstance(value, (list, tuple)):
                return be_type(*value)
        return value

    def api_coerce_value(self, attr, value):
        """Perform type checking of value.

        Will coerce value into expected type if needed and possible.

        Args:
            attr (:obj:`str`):
                Attribute that is being set with value on obj.
            value (:obj:`object`):
                Value of attr being checked.

        Raises:
            :exc:`tantrum.api_models.exceptions.AttrTypeError`:
                If value is not the expected type after all of this magic.
            :exc:`tantrum.api_models.exceptions.AttrUndefinedError`:
                If value type is not :data:`simple_types` and attr is not in
                :attr:`api_attrs`.
            :exc:`tantrum.api_models.exceptions.AttrUndefinedWarning`:
                If value type is :data:`simple_types` and attr not in
                :attr:`api_attrs`.

        Notes:
            No checking or coercion is done if:
                * value is None
                * attr starts with "_"
                * attr is all upper case

        Returns:
            :obj:`object`

        """
        if value is None or attr.startswith("_") or attr.isupper():
            return value

        if attr not in self.api_attrs():
            if isinstance(value, bool) or isinstance(value, integer_types):
                self.API_SIMPLE[attr] = integer_types
            elif isinstance(value, float_types):
                self.API_SIMPLE[attr] = float_types
            elif isinstance(value, string_types):
                self.API_SIMPLE[attr] = string_types
            else:
                raise exceptions.AttrUndefinedError(obj=self, attr=attr, value=value)
            warnings.warn(
                exceptions.AttrUndefinedWarning(obj=self, attr=attr, value=value)
            )

        be_type = self.api_attrs()[attr]

        if attr in self.API_SIMPLE:
            value = self.api_coerce_simple(value=value, be_type=be_type)
        if attr in self.API_COMPLEX:
            value = self.api_coerce_complex(value=value, be_type=be_type)

        if isinstance(value, be_type) or value is None:
            return value
        raise exceptions.AttrTypeError(
            obj=self, value=value, attr=attr, be_type=be_type
        )

    def serialize_json(self):
        ser = self.serialize()
        return json.dumps(ser, indent=2)


class ApiItem(ApiModel):
    """Model for a complex item in the API."""

    API_LIST_API_NAME = None
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""

    API_LIST_CLS = None
    """:class:`ApiList`: List class that holds this item class."""

    def __init__(self, **kwargs):
        """Constructor.

        Args:
            **kwargs:
                Set and checked using :meth:`ApiModel.api_attrs`.

        """
        for attr in self.api_attrs():
            setattr(self, attr, kwargs.pop(attr, None))
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def __len__(self):
        """Return number of :meth:`ApiModel.api_attrs` that are not None.

        Returns:
            :obj:`int`

        """
        api_attrs = self.api_attrs()
        set_attrs = [k for k in api_attrs if getattr(self, k, None) is not None]
        return len(set_attrs)

    def __str__(self):
        """Show object info using :meth:`ApiModel.api_attrs_str`.

        Returns:
            :obj:`str`

        """
        show_attrs = self.api_attrs_str()
        attrs = []
        for attr in show_attrs:
            value = getattr(self, attr, None)
            tmpl = self.T_ATTR_STR if attr in self.API_COMPLEX else self.T_ATTR_REPR
            attrs.append(tmpl.format(attr=attr, value=value))

        attrs = self.T_ATTR_JOIN.join(attrs)
        return self.T_CLS_STR.format(obj=self, attrs=attrs)

    def __repr__(self):
        """Show object info using :meth:`ApiModel.api_attrs_repr`.

        Returns:
            :obj:`str`

        """
        show_attrs = self.api_attrs_repr()
        attrs = []
        for attr in show_attrs:
            value = getattr(self, attr, None)
            tmpl = self.T_ATTR_REPR
            attrs.append(tmpl.format(attr=attr, value=value))
        attrs = self.T_ATTR_JOIN.join(attrs)
        return self.T_CLS_REPR.format(obj=self, attrs=attrs)

    def __setattr__(self, attr, value):
        """Enforce type using :meth:`ApiModel.api_coerce_value`."""
        value = self.api_coerce_value(attr=attr, value=value)
        super(ApiItem, self).__setattr__(attr, value)

    def serialize(
        self,
        empty=False,
        list_attrs=False,
        exclude_attrs=None,
        only_attrs=None,
        wrap_name=True,
        wrap_item_attr=True,
    ):
        """Serialize this object into a dict.

        Args:
            empty (:obj:`bool`, optional):
                Include attrs that have a value of None when serializing.

                Defaults to: False.
            list_attrs (:obj:`bool`, optional):
                Include simple attrs of :obj:`ApiList` when serializing.

                Defaults to: False.
            exclude_attrs (:obj:`list` of :obj:`str`, optional):
                Exclude these attrs when serializing.

                Defaults to: None.
            only_attrs (:obj:`list` of :obj:`str`, optional):
                Include only these attrs when serializing.

                Defaults to: None.
            wrap_name (:obj:`bool`, optional):
                Wrap return in a dict with key of :attr:`ApiModel.API_NAME` and
                value of return.

                Defaults to: True.
            wrap_item_attr (:obj:`bool`, optional):
                Wrap return in a dict with key of :attr:`ApiList.API_ITEM_ATTR`
                and value of return.

                Only used when serializing items of type :obj:`ApiList`.

                Defaults to: True.

        Notes:
            All child objects will get wrap_name=False.

        Returns:
            :obj:`dict`

        """
        exclude_attrs = self.api_coerce_list(value=exclude_attrs)
        only_attrs = self.api_coerce_list(value=only_attrs)
        ret = {}
        sargs = {
            "empty": empty,
            "exclude_attrs": exclude_attrs,
            "only_attrs": only_attrs,
            "list_attrs": list_attrs,
            "wrap_name": False,
            "wrap_item_attr": wrap_item_attr,
        }
        for attr in self.api_attrs():
            if attr in exclude_attrs or (only_attrs and attr not in only_attrs):
                continue
            value = getattr(self, attr, None)
            is_model = isinstance(value, ApiModel)
            value = value.serialize(**sargs) if is_model else value
            include = (value is None and empty) or value is not None
            ret.update({attr: value} if include else {})
        if wrap_name:
            ret = {self.API_NAME: ret}
        return ret


class ApiList(ApiModel):
    """Model for an array in the API."""

    API_ITEM_ATTR = None
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = None
    """:class:`ApiItem`: Item class this list class holds."""

    LIST = None
    """:obj:`list` of :attr:`API_ITEM_CLS`: List container for this class."""

    T_CLS_STR = "{obj.__class__.__name__}({attrs}) with {count} {item_cls} objects"
    """:obj:`str`: Template for class name and attrs in str."""

    def __init__(self, *args, **kwargs):
        """Constructor.

        Args:
            *args:
                Items of type :attr:`API_ITEM_CLS`:
                    Will be added to :attr:`LIST`.
            **kwargs:
                attr from :attr:`API_ITEM_ATTR` (:obj:`list`):
                    List of :attr:`API_ITEM_CLS` to add to :attr:`LIST`.
                rest of kwargs:
                    Set on this object with type checking using.

        """
        super(ApiList, self).__setattr__("LIST", [])

        # support obj(item1, item2, **kwargs)
        items = list(args)

        # support obj(item_attr=value)
        if self.API_ITEM_ATTR in kwargs:
            value = kwargs.pop(self.API_ITEM_ATTR)
            # support obj(self.API_ITEM_ATTR=(item1, item2))
            if isinstance(value, tuple):
                value = list(value)
            # support obj(self.API_ITEM_ATTR=None)
            if value is None:
                value = []
            # support obj(self.API_ITEM_ATTR=item1)
            if not isinstance(value, list):
                value = [value]
            # finally, support obj(self.API_ITEM_ATTR=[item1, item2])
            items += value

        self.LIST = items

        for attr in self.api_attrs():
            setattr(self, attr, kwargs.pop(attr, None))
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def __str__(self):
        """Show object info using :meth:`ApiModel.api_attrs_str`.

        Returns:
            :obj:`str`

        """
        show_attrs = self.api_attrs_str()
        attrs = []
        for attr in show_attrs:
            value = getattr(self, attr, None)
            tmpl = self.T_ATTR_STR if attr in self.API_COMPLEX else self.T_ATTR_REPR
            attrs.append(tmpl.format(attr=attr, value=value))

        attrs = self.T_ATTR_JOIN.join(attrs)
        return self.T_CLS_STR.format(
            obj=self, attrs=attrs, count=len(self), item_cls=self.api_item_cls_str()
        )

    def __repr__(self):
        """Show object info using :meth:`ApiModel.api_attrs_repr`.

        Returns:
            :obj:`str`

        """
        show_attrs = self.api_attrs_repr()
        attrs = []
        tmpl = self.T_ATTR_REPR
        for attr in show_attrs:
            value = getattr(self, attr, None)
            attrs.append(tmpl.format(attr=attr, value=value))
        value = [format(i) for i in self]
        attrs.append(tmpl.format(attr=self.API_ITEM_ATTR, value=value))
        attrs = self.T_ATTR_JOIN.join(attrs)
        return self.T_CLS_REPR.format(obj=self, attrs=attrs)

    def __len__(self):
        """Return length of list container :attr:`LIST`.

        Returns:
            :obj:`int`

        """
        return len(self.LIST)

    def __contains__(self, value):
        """Support in operand.

        Args:
            value (:obj:`object`):
                Check if this value in :attr:`LIST`.

        Notes:
            Will call :meth:`api_coerce_item` to check value is of type defined in
            :attr:`API_ITEM_CLS`, but will return False if value is of wrong type
            and can not be coerced.

        Returns:
            :obj:`bool`:

        """
        try:
            value = self.api_coerce_item(item=value, items=None, attr=None, op="in")
        except Exception:
            return False
        return value in self.LIST

    def __add__(self, value):
        """Support + operand.

        Args:
            value (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                ApiList of this same type or a python list type to combine
                with this object.

        Notes:
            This will use :meth:`api_coerce_items` to perform type checking
            on each item in value.

        Returns:
            :obj:`ApiList`:
                A new instance with list items from this object combined with
                list items from value.

        """
        api_attrs = self.api_attrs()

        val_items = self.api_coerce_items(attr=None, value=value, op="+=")
        mod_items = val_items + self.LIST
        attrs = {k: getattr(self, k, None) for k, v in api_attrs.items()}
        return self.__class__(*mod_items, **attrs)

    def __iadd__(self, value):
        """Support += operand.

        Args:
            value (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                ApiList of this same type or a python list type to combine
                with this object.

        Notes:
            This will use :meth:`api_coerce_items` to perform type checking
            on each item in value.

        Returns:
            :obj:`ApiList`:
                This object with list items from value appended.

        """
        mod_items = self.api_coerce_items(attr=None, value=value, op="+=")
        self.LIST += mod_items
        return self

    def __getitem__(self, value):
        """Support indexing of list container.

        Args:
            value (:obj:`int`):
                Index of item to retrieve from :attr:`LIST`.

        Returns:
            :attr:`API_ITEM_CLS`:
                The type of object this list contains.

        """
        return self.LIST[value]

    def __setattr__(self, attr, value):
        """Enforce type checking for attr and value.

        Args:
            attr (:obj:`str`):
                Attribute to set on this object.
            value (:obj:`object`):
                Value to perform type checking on using
                :func:`ApiModel.api_coerce_value`.

        Notes:
            If attr is either "LIST" or :attr:`API_ITEM_ATTR`, will use
            :meth:`api_coerce_items` to perform type checking on each item in value.

            If attr is same as :attr:`API_ITEM_ATTR`, the attr will
            be changed to "LIST".

        """
        if attr in [self.API_ITEM_ATTR, "LIST"]:
            value = self.api_coerce_items(attr=attr, value=value, op="=")
            attr = "LIST"
        else:
            value = self.api_coerce_value(attr=attr, value=value)
        super(ApiList, self).__setattr__(attr, value)

    def api_coerce_items(self, attr, value, op):
        """Check that value is a list type and that all items value are the proper type.

        Args:
            attr (:obj:`str`):
                Attribute being set on this object.
            value (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                The list object holding items that should have type checking.
            op (:obj:`str`):
                Operation being performed that called this method.

                Used in exception messages.

        Raises:
            :exc:`tantrum.api_models.exceptions.ListTypeError`:
                If value is not a :obj:`list` or :obj:`tuple` or an :obj:`ApiList`
                that is the same type as this objects class.

        Notes:
            Will call :meth:`api_coerce_items_hook` before checking each item
            in value, which can be overridden by subclasses to modify value.

            Will call :meth:`api_coerce_item` to check each item in value
            is of type defined in :attr:`API_ITEM_CLS`.

        Returns:
            :obj:`list`:
                The list of items with their values checked and coerced.

        """
        if value is None:
            return []
        is_list = isinstance(value, (list, tuple))
        is_same = isinstance(value, self.__class__)
        if not any([is_list, is_same]):
            raise exceptions.ListTypeError(obj=self, attr=attr, value=value, op=op)
        value = self.api_coerce_items_hook(attr=attr, value=value, op=op)
        return [
            self.api_coerce_item(item=i, items=value, attr=attr, op=op) for i in value
        ]

    def api_coerce_items_hook(self, attr, value, op):
        """Check hook that allows subclasses to modify list items.

        Args:
            attr (:obj:`str`):
                Attribute being set on this object.
            value (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                The list object holding items.
            op (:obj:`str`):
                Operation being performed that called this method.

        Notes:
            By default, this does nothing but return value unchanged.

        Returns:
            :obj:`list`:
                The list of items modified or as is.

        """
        return value

    def api_coerce_item(self, item, items, attr, op):
        """Perform type checking of a list item.

        Args:
            item (:obj:`object`):
                Item that is being checked.
            items (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                The list object that item came from.
            op (:obj:`str`):
                Operation being performed that called this method.

                Used in exception messages.

        Notes:
            Will coerce item into expected type if needed and possible.

        Raises:
            :exc:`tantrum.api_models.exceptions.ListItemTypeError`:
                If item does not match the type from :obj:`API_ITEM_CLS`
                and it can not be coerced into the expected type.

        Returns:
            :obj:`object`

        """
        be_type = self.API_ITEM_CLS

        if be_type is None:
            return item

        if self.api_item_cls_is_complex():
            if isinstance(item, dict):
                return be_type(**item)
            elif isinstance(item, (list, tuple)):
                return be_type(*item)

        if not isinstance(item, be_type):
            if be_type == float_types:
                return self.api_coerce_float(value=item)
            if be_type == integer_types:
                return self.api_coerce_int(value=item)
            if be_type == string_types and isinstance(item, simple_types):
                return item

        if not isinstance(item, be_type):
            raise exceptions.ListItemTypeError(
                obj=self, item=item, items=items, op=op, attr=attr, be_type=be_type
            )
        return item

    @classmethod
    def api_attrs_desc(cls):
        """Get simple and complex attributes of an ApiModel object in string format.

        Notes:
            Used by exceptions to add defined attributes to error messages.

        Returns:
            :obj:`list` of :obj:`str`

        """
        lines = super(ApiList, cls).api_attrs_desc()
        tmpl = cls.T_ATTR_DESC.format
        desc = tmpl(api_type="List", attr=cls.API_ITEM_ATTR, attr_type=cls.API_ITEM_CLS)
        lines.append(desc)
        return lines

    @classmethod
    def api_item_cls_str(cls):
        """Get str of :attr:`API_ITEM_CLS`.

        Returns:
            :obj:`str`

        """
        classes = cls.api_coerce_list(cls.API_ITEM_CLS)
        names = [x.__name__ for x in classes if hasattr(x, "__name__")]
        names = names or ["ANY TYPE"]
        return ", ".join(names)

    @classmethod
    def api_item_cls_is_complex(cls):
        """Check if :attr:`API_ITEM_CLS` is an ApiModel sub class.

        Returns:
            :obj:`bool`

        """
        check = cls.api_coerce_list(cls.API_ITEM_CLS)
        return any(issubclass(x, (ApiItem, ApiList)) for x in check)

    def append(self, value):
        """Support appending an item to list container.

        Args:
            value (:obj:`object`):
                Item to append to :attr:`LIST`.

        Notes:
            Will call :meth:`api_coerce_item` to check value
            is of type defined in :attr:`API_ITEM_CLS`.

        """
        value = self.api_coerce_item(item=value, items=None, attr=None, op="append()")
        self.LIST.append(value)

    def remove(self, value):
        """Support removing an item from list container.

        Args:
            value (:obj:`object`):
                Item to remove from :attr:`LIST`.

        """
        value = self.api_coerce_item(item=value, items=None, attr=None, op="remove()")
        return self.LIST.remove(value)

    def pop(self, value=-1):
        """Support pop of an item from list container.

        Args:
            value (:obj:`object`, optional):
                Item to pop from :attr:`LIST`.

                Defaults to: -1.

        """
        return self.LIST.pop(value)

    def reverse(self):
        """Support reverse on list container."""
        self.LIST.reverse()

    def get_item_by_attr(self, value, attr="name", regex_value=False):
        """Support getting an item from list container by attr value.

        Args:
            value (:obj:`object`):
                Value to check against the attr value of each item in :attr:`LIST`.
            attr (:obj:`str`, optional):
                Attribute to check if value matches.

                Defaults to: "name".
            regex_value (:obj:`bool`, optional):
                Treat value as a regex pattern instead of comparing equality of value
                to attr value.

                Defaults to: False.

        Raises:
            :exc:`tantrum.api_models.exceptions.GetSingleItemError`:
                If number of items in :attr:`LIST` whose attr value
                matches value is either 0 or more than 1.

        Returns:
            :obj:`object`:
                The object from :attr:`LIST` whose attr value matches value.

        """
        items = []
        for i in self.LIST:
            ival = getattr(i, attr, None)
            if regex_value and re.search(format(value), format(ival)):
                items.append(i)
            elif ival == value:
                items.append(i)
        if len(items) == 1:
            return items[0]
        raise exceptions.GetSingleItemError(
            obj=self, value=value, attr=attr, regex_value=regex_value, items=items
        )

    def get_items_by_attr(self, value, attr="name", regex_value=False, new_list=False):
        """Support getting items from list container by attr value.

        Args:
            value (:obj:`object`):
                Value to check against the attr value of each item in :attr:`LIST`.
            attr (:obj:`str`, optional):
                Attribute to check if value matches.

                Defaults to: "name".
            regex_value (:obj:`bool`, optional):
                Treat value as a regex pattern instead of comparing equality of value
                to attr value.

                Defaults to: False.
            new_list (:obj:`bool`, optional):
                Return a new class of this objects type with matching items instead
                of a regular python list type.

                Defaults to: False.

        Returns:
            :obj:`list` of :obj:`object`:
                All objects from :attr:`LIST` whose attr value matches value.

        """
        items = []
        for i in self.LIST:
            ival = getattr(i, attr, None)
            if regex_value and re.search(format(value), format(ival)):
                items.append(i)
            elif ival == value:
                items.append(i)
        return self.__class__(*items) if new_list else items

    def pop_item_by_attr(self, value, attr="name", regex_value=False):
        """Support popping an item from list container by attr value.

        Args:
            value (:obj:`object`):
                Value to check against the attr value of each item in :attr:`LIST`.
            attr (:obj:`str`, optional):
                Attribute to check if value matches.

                Defaults to: "name".
            regex_value (:obj:`bool`, optional):
                Treat value as a regex pattern instead of comparing equality of value
                to attr value.

                Defaults to: False.

        Raises:
            :exc:`tantrum.api_models.exceptions.GetSingleItemError`:
                If number of items in :attr:`LIST` whose attr value
                matches value is either 0 or more than 1.

        Returns:
            :obj:`object`:
                The object popped from :attr:`LIST` whose attr value matches value.

        """
        item = self.get_item_by_attr(value, attr, regex_value)
        return self.LIST.pop(self.LIST.index(item))

    def pop_items_by_attr(self, value, attr="name", regex_value=False, new_list=False):
        """Support popping items from list container by attr value.

        Args:
            value (:obj:`object`):
                Value to check against the attr value of each item in :attr:`LIST`.
            attr (:obj:`str`, optional):
                Attribute to check if value matches.

                Defaults to: "name".
            regex_value (:obj:`bool`, optional):
                Treat value as a regex pattern instead of comparing equality of value
                to attr value.

                Defaults to: False.
            new_list (:obj:`bool`, optional):
                Return a new class of this objects type with matching items instead
                of a regular python list type.

                Defaults to: False.

        Returns:
            :obj:`list` of :obj:`object`:
                All objects popped from :attr:`LIST` whose attr value matches value.

        """
        items = self.get_items_by_attr(value, attr, regex_value, new_list)
        for i in items:
            self.LIST.remove(i)
        return items

    def sort(self, key=operator.attrgetter("id"), reverse=False):
        """Support sorting items in list container in place.

        Args:
            key (:obj:`object`, optional):
                Key to sort items in :attr:`LIST` on.

                Defaults to: operator.attrgetter("id").
            reverse (:obj:`bool`, optional):
                Reverse the sort mechanism.

                Defaults to: False.

        """
        self.LIST.sort(key=key, reverse=reverse)

    def serialize(
        self,
        empty=False,
        list_attrs=False,
        exclude_attrs=None,
        only_attrs=None,
        wrap_name=True,
        wrap_item_attr=True,
    ):
        """Serialize this object into a dict.

        Args:
            empty (:obj:`bool`, optional):
                Include attrs that have a value of None when serializing.

                Defaults to: False.
            list_attrs (:obj:`bool`, optional):
                Include simple attrs of :obj:`ApiList` when serializing.

                Defaults to: False.
            exclude_attrs (:obj:`list` of :obj:`str`, optional):
                Exclude these attrs when serializing.

                Defaults to: None.
            only_attrs (:obj:`list` of :obj:`str`, optional):
                Include only these attrs when serializing.

                Defaults to: None.
            wrap_name (:obj:`bool`, optional):
                Wrap return in a dict with key of :attr:`ApiModel.API_NAME` and
                value of return.

                Defaults to: True.
            wrap_item_attr (:obj:`bool`, optional):
                Wrap return in a dict with key of :attr:`ApiList.API_ITEM_ATTR`
                and value of return.

                Only used when serializing items of type :obj:`ApiList`.

                Defaults to: True.

        Notes:
            All child objects will get wrap_name=False.

            If wrap_name and wrap_item_attr are both False,
            return will be :obj:`list` of :obj:`dict`, otherwise return will be a
            :obj:`dict`.

        Returns:
            :obj:`list` of :obj:`object` or :obj:`dict`

        """
        exclude_attrs = self.api_coerce_list(value=exclude_attrs)
        only_attrs = self.api_coerce_list(value=only_attrs)
        sargs = {
            "empty": empty,
            "exclude_attrs": exclude_attrs,
            "only_attrs": only_attrs,
            "list_attrs": list_attrs,
            "wrap_name": False,
            "wrap_item_attr": wrap_item_attr,
        }
        simples = {}
        if list_attrs:
            for attr in self.api_attrs():
                if attr in exclude_attrs or (only_attrs and attr not in only_attrs):
                    continue
                value = getattr(self, attr, None)
                is_model = isinstance(value, ApiModel)
                value = value.serialize(**sargs) if is_model else value
                include = (value is None and empty) or value is not None
                simples.update({attr: value} if include else {})
        items = [i.serialize(**sargs) if isinstance(i, ApiModel) else i for i in self]
        if wrap_item_attr:
            ret = {}
            ret.update(simples)
            sub_ret = {self.API_ITEM_ATTR: items}
            if wrap_name:
                sub_ret = {self.API_NAME: sub_ret}
            ret.update(sub_ret)
            return ret
        if wrap_name:
            ret = {}
            ret.update(simples)
            ret.update({self.API_NAME: items})
            return ret
        return items



# -*- coding: utf-8 -*-
"""Client for making HTTP requests."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests

from . import exceptions
from .. import utils
from .. import version


class HttpClient(object):
    """Convenience class for requests package."""

    def __init__(
        self,
        url,
        verify=True,
        timeout=5,
        save_last=False,
        save_history=False,
        log_request=False,
        log_response=False,
        lvl="info",
    ):
        """Constructor.

        Args:
            url (:obj:`str`):
                URL to use.
            timeout (:obj:`int`, optional):
                Connect timeout to use for requests.

                Defaults to: 5.
            lvl (:obj:`str`, optional):
                Logging level for this object.

                Defaults to: "info".
            verify (:obj:`bool` or :obj:`str`, optional):
                Enable/Disable SSL certificate validation using built in CAs,
                or a path to custom cert.

                Defaults to: True.
            save_last (:obj:`bool`, optional):
                Save last request to :attr:`HttpClient.last_request` and last response
                to :attr:`HttpClient.last_response`.

                Defaults to: False.
            save_history (:obj:`bool`, optional):
                Append last response to :attr:`HttpClient.history`.

                Defaults to: False.
            log_request (:obj:`bool`, optional):
                Log request details to debug level.

                Defaults to: False.
            log_response (:obj:`bool`, optional):
                Log response details to debug level.

                Defaults to: False.

        Notes:
            If verify is True or None, verification is done using default/built in
            CA. OS env vars $REQUESTS_CA_BUNDLE, and $CURL_CA_BUNDLE are used if set
            and trust_env is True, and if trust_env is False session's verify is used.

            If verify is False, no verification is done. This overrides OS env and
            session's verify for this request and no verification is done at all.
            Don't do this.

            If verify is a str, verification is done with PEM file at path.
            This overrides OS env and session's verify for this request.

            Caveat: If previous request made with session and close has not been
            called on session, the verify of the previous request will be used
            no matter what is supplied here.

        """
        self.log = utils.logs.get_obj_log(obj=self, lvl=lvl)
        """:obj:`logging.Logger`: Log for this object."""
        self.timeout = timeout
        """:obj:`int`: Connect timeout used for all requests."""
        self.parsed_url = self.parse_url(url)
        """:obj:`UrlParser`: Parsed version of URL."""
        self.last_request = None
        """:obj:`requests.PreparedRequest`: Last request sent."""
        self.last_response = None
        """:obj:`requests.Response`: Last response received."""
        self.history = []
        """:obj:`list`: History of all responses received."""
        self.verify = verify
        """:obj:`bool`: SSL Verification."""
        self.save_last = save_last
        """:obj:`bool`: Save requests/responses to last_request/last_response."""
        self.save_history = save_history
        """:obj:`bool`: Append requests/responses to history."""
        self.log_request = log_request
        """:obj:`bool`: Log requests."""
        self.log_response = log_response
        """:obj:`bool`: Log responses."""
        self.session = requests.Session()
        """:obj:`requests.Session`: Requests session object."""

    @property
    def url(self):
        """Get the URL string from :attr:`HttpClient.parsed_url`.

        Returns:
            :obj:`str`

        """
        return self.parsed_url.url

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = ["url={!r}".format(self.url)]
        bits = "({})".format(", ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    def __call__(
        self,
        method="get",
        path="",
        data=None,
        timeout=5,
        params=None,
        headers=None,
        b64_headers=None,
        verify=None,
        save_last=None,
        save_history=None,
        log_request=None,
        log_response=None,
        cause="",
    ):
        """Create, prepare a request, and then send it.

        Args:
            method (:obj:`str`, optional):
                Method to use.

                Defaults to: "get".
            path (:obj:`str`, optional):
                Path to append to :attr:`HttpClient.url` for this request.

                Defaults to: "".
            data (:obj:`str`, optional):
                Data to send with POST.

                Defaults to: None.
            timeout (:obj:`int`, optional):
                Response timeout.

                Defaults to: 5.
            params (:obj:`dict`, optional):
                URL parameters.

                Defaults to: None.
            headers (:obj:`dict`, optional):
                Headers.

                Defaults to: None.
            b64_headers (:obj:`list` of :obj:`str`, optional):
                Headers to base 64 encode.

                Defaults to: None.
            verify (:obj:`bool` or :obj:`str`, optional):
                Enable SSL certification validation.
                If None uses :attr:`HttpClient.verify`.

                Defaults to: None.
            save_last (:obj:`bool`, optional):
                Save last request to :attr:`HttpClient.last_request` and last response
                to :attr:`HttpClient.last_response`.
                If None uses :attr:`HttpClient.save_last`.

                Defaults to: None.
            save_history (:obj:`bool`, optional):
                Append last response to :attr:`HttpClient.history`.
                If None uses :attr:`HttpClient.save_history`.

                Defaults to: None.
            log_request (:obj:`bool`, optional):
                Log request details to debug level.
                If None uses :attr:`HttpClient.log_request`.

                Defaults to: None.
            log_response (:obj:`bool`, optional):
                Log response details to debug level.
                If None uses :attr:`HttpClient.log_response`.

                Defaults to: None.
            cause (:obj:`str`, optional):
                String to explain purpose of request.

                Defaults to: "".

        Returns:
            :obj:`requests.Response`

        """
        b64 = b64_headers or []
        headers = headers or {}
        headers = {k: v for k, v in headers.items()}

        verify = utils.tools.def_none(verify, self.verify)
        save_last = utils.tools.def_none(save_last, self.save_last)
        save_history = utils.tools.def_none(save_history, self.save_history)
        log_request = utils.tools.def_none(log_request, self.log_request)
        log_response = utils.tools.def_none(log_response, self.log_response)

        h = {}
        h.update(headers)
        h.setdefault("User-Agent", self.user_agent)
        h.update({k: utils.tools.b64_encode(v) for k, v in h.items() if k in b64})

        req_args = {}
        req_args["url"] = requests.compat.urljoin(self.url, path)
        req_args["method"] = method
        req_args["data"] = data
        req_args["headers"] = h
        req_args["params"] = params

        request = requests.Request(**req_args)

        prequest = self.session.prepare_request(request=request)
        prequest.cause = cause

        if save_last:
            self.last_request = prequest

        if log_request:
            m = ["request sent: url={p.url!r}", "method={p.method!r}", "size={size}"]
            m = ", ".join(m)
            m = m.format(p=prequest, size=len(prequest.body or ""))
            self.log.debug(m)

        send_args = {}
        send_args["timeout"] = (self.timeout, timeout)
        send_args.update(
            self.session.merge_environment_settings(
                url=self.url,
                proxies=None,  # rely on OS env proxies, then self.session.proxies
                stream=None,  # not using
                verify=verify,
                cert=None,  # rely on client cert set in self.session.cert
            )
        )

        r = self.session.send(prequest, **send_args)
        r.cause = cause

        if save_last:
            self.last_response = r

        if save_history:
            self.history = getattr(self, "history", [])
            self.history.append(r)

        if log_response:
            m = [
                "response received: url={r.url!r}",
                "method={r.request.method!r}",
                "size={size}",
                "status={r.status_code!r}",
                "reason={r.reason!r}",
                "elapsed={r.elapsed}",
                "cause={r.cause!r}",
            ]
            m = ", ".join(m)
            m = m.format(r=r, size=len(r.text or ""))
            self.log.debug(m)

        return r

    def parse_url(self, url):
        """Parse a URL using UrlParser.

        Args:
            url (:obj:`str`):
                URL to parse.

        Returns:
            :obj:`UrlParser`

        """
        parsed_url = UrlParser(url=url, default_scheme="https")
        m = "Parsed url {old!r} into {new!r} using {parsed}"
        m = m.format(old=url, new=parsed_url.url, parsed=parsed_url)
        self.log.debug(m)
        return parsed_url

    @property
    def user_agent(self):
        """Build a user agent string for use in headers.

        Returns:
            :obj:`str`

        """
        return "{pkg}.{name}/{ver}".format(
            pkg=__name__.split(".")[0],
            name=self.__class__.__name__,
            ver=version.__version__,
        )


class UrlParser(object):
    """Parse a URL and ensure it has the neccessary bits."""

    def __init__(self, url, default_scheme=""):
        """Constructor.

        Args:
            url (:obj:`str`):
                URL to parse
            default_scheme (:obj:`str`, optional):
                If no scheme in URL, use this.

                Defaults to: ""

        Raises:
            :exc:`exceptions.ModuleError`:
                If parsed URL winds up without a hostname, port, or scheme.

        """
        self._init_url = url
        """:obj:`str`: Initial URL provided."""
        self._init_scheme = default_scheme
        """:obj:`str`: Default scheme provided."""
        self._init_parsed = requests.compat.urlparse(url)
        """:obj:`urllib.parse.ParseResult`: First pass of parsing URL."""
        self.parsed = self.reparse(
            parsed=self._init_parsed, default_scheme=default_scheme
        )
        """:obj:`urllib.parse.ParseResult`: Second pass of parsing URL."""

        for part in ["hostname", "port", "scheme"]:
            if not getattr(self.parsed, part, None):
                error = "\n".join(
                    [
                        "",
                        "Parsed into: {pstr}",
                        "URL format should be like: scheme://hostname:port",
                        "No {part} provided in URL {url!r}",
                    ]
                )
                error = error.format(part=part, url=url, pstr=self.parsed_str)
                raise exceptions.ModuleError(error)

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = ["parsed={!r}".format(self.parsed_str)]
        bits = "({})".format(", ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @property
    def hostname(self):
        """Hostname part from :attr:`UrlParser.parsed`.

        Returns:
            :obj:`str`

        """
        return self.parsed.hostname

    @property
    def port(self):
        """Port part from :attr:`UrlParser.parsed`.

        Returns:
            :obj:`int`

        """
        return int(self.parsed.port)

    @property
    def scheme(self):
        """Scheme part from :attr:`UrlParser.parsed`.

        Returns:
            :obj:`str`

        """
        return self.parsed.scheme

    @property
    def url(self):
        """Get scheme, hostname, and port from :attr:`UrlParser.parsed`.

        Returns:
            :obj:`str`

        """
        return self.unparse_base(p=self.parsed)

    @property
    def url_full(self):
        """Get full URL from :attr:`UrlParser.parsed`.

        Returns:
            :obj:`str`

        """
        return self.unparse_all(p=self.parsed)

    @property
    def parsed_str(self):
        """Create string of :attr:`UrlParser.parsed`.

        Returns:
            :obj:`str`

        """
        parsed = getattr(self, "parsed", None)
        attrs = [
            "scheme",
            "netloc",
            "hostname",
            "port",
            "path",
            "params",
            "query",
            "fragment",
        ]
        vals = ", ".join(
            [
                "{a}={v!r}".format(a=a, v="{}".format(getattr(parsed, a, "")) or "")
                for a in attrs
            ]
        )
        return vals

    def make_netloc(self, host, port):
        """Create netloc from host and port.

        Args:
            host (:obj:`str`):
                Host part to use in netloc.
            port (:obj:`str`):
                Port part to use in netloc.

        Returns:
            :obj:`str`

        """
        netloc = ":".join([host, port]) if port else host
        return netloc

    def reparse(self, parsed, default_scheme=""):
        """Reparse a parsed URL into a parsed URL with values fixed.

        Args:
            parsed (:obj:`urllib.parse.ParseResult`):
                Parsed URL to reparse.
            default_scheme (:obj:`str`, optional):
                If no scheme in URL, use this.

                Defaults to: ""

        Returns:
            :obj:`urllib.parse.ParseResult`

        """
        scheme, netloc, path, params, query, fragment = parsed
        host = parsed.hostname
        port = format(parsed.port or "")

        if not netloc and scheme and path and path.split("/")[0].isdigit():
            """For case:
            >>> urllib.parse.urlparse('host:443/')
            ParseResult(
                scheme='host', netloc='', path='443/', params='', query='', fragment=''
            )
            """
            host = scheme  # switch host from scheme to host
            port = path.split("/")[0]  # remove / from path and assign to port
            path = ""  # empty out path
            scheme = default_scheme
            netloc = ":".join([host, port])

        if not netloc and path:
            """For cases:
            >>> urllib.parse.urlparse('host:443')
            ParseResult(
                scheme='', netloc='', path='host:443', params='', query='', fragment=''
            )
            >>> urllib.parse.urlparse('host')
            ParseResult(
                scheme='', netloc='', path='host', params='', query='', fragment=''
            )
            """
            netloc, path = path, netloc
            if ":" in netloc:
                host, port = netloc.split(":", 1)
                netloc = ":".join([host, port]) if port else host
            else:
                host = netloc

        scheme = scheme or default_scheme
        if not scheme and port:
            if format(port) == "443":
                scheme = "https"
            elif format(port) == "80":
                scheme = "http"

        if not port:
            if scheme == "https":
                port = "443"
                netloc = self.make_netloc(host, port)
            elif scheme == "http":
                port = "80"
                netloc = self.make_netloc(host, port)

        pass2 = requests.compat.urlunparse(
            (scheme, netloc, path, params, query, fragment)
        )
        ret = requests.compat.urlparse(pass2)
        return ret

    def unparse_base(self, p):
        """Unparse a parsed URL into just the scheme, hostname, and port parts.

        Args:
            p (:obj:`urllib.parse.ParseResult`):
                Parsed URL to unparse.

        Returns:
            :obj:`str`

        """
        # only unparse self.parsed into url with scheme and netloc
        return requests.compat.urlunparse((p.scheme, p.netloc, "", "", "", ""))

    def unparse_all(self, p):
        """Unparse a parsed URL with all the parts.

        Args:
            p (:obj:`urllib.parse.ParseResult`):
                Parsed URL to unparse.

        Returns:
            :obj:`str`

        """
        return requests.compat.urlunparse(p)



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.api_objects`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


class ModuleError(exceptions.PackageError):
    """Parent of all exceptions for :mod:`tantrum.api_objects`."""

    pass


class ModuleWarning(exceptions.PackageWarning):
    """Parent of all warnings for :mod:`tantrum.api_objects`."""

    pass


class NoVersionFoundError(ModuleError):
    """Exception handler when finding a version of an API object module."""

    pass


class UnknownApiNameError(ModuleError):
    """Exception handler when unable to find an API name for an API object."""

    def __init__(self, name, name_map, module):
        """Constructor.

        Args:
            name (:obj:`str`):
                API name being searched for.
            name_map (:obj:`dict`):
                Map of all API names to API objects.
            module (:obj:`object`):
                Source API module of name_map.

        """
        self.name = name
        """:obj:`str`: API name being searched for."""
        self.name_map = name_map
        """:obj:`dict`: Map of all API names to API objects."""
        self.module = module
        """:obj:`object`: Source API module of name_map."""

        error = [
            "API names of all API objects:{valid_names}",
            "API module: {module}",
            "Unable to find a matching API class for API name {name!r}",
        ]
        error = "\n".join(error)
        self.error = error.format(
            name=name,
            module=module,
            valid_names="\n  - " + "\n  - ".join(sorted(list(name_map.keys()))),
        )
        """:obj:`str`: Error message that was thrown."""
        super(UnknownApiNameError, self).__init__(self.error)



# -*- coding: utf-8 -*-
"""Result objects that deserialize responses from Tanium API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import abc
import re
import six

import xmltodict

from . import exceptions
from .. import utils


@six.add_metaclass(abc.ABCMeta)
class Result(object):
    """Deserialize API responses from Tanium API."""

    @classmethod
    @abc.abstractmethod
    def from_response(cls, api_objects, response, lvl="info"):
        """Create :obj:`Result` from :obj:`requests.Response`.

        Args:
            api_objects (:obj:`tantrum.api_objects.ApiObjects`):
                API Objects Container to use.
            response (:obj:`requests.Response`):
                Response object received from Tanium API request.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        Returns:
            :obj:`Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def error_check(self):
        """Check for errors in :attr:`response_body_str`.

        Raises:
            :exc:`exceptions.ObjectExistsError`
            :exc:`exceptions.ObjectNotFoundError`
            :exc:`exceptions.ResponseError`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def error_text(self):
        """Get the error text from :attr:`response_body_str`.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def api_objects(self):
        """Get the API objects container.

        Returns:
            :obj:`tantrum.api_objects.ApiObjects`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def url(self):
        """Get the URL used in request.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def status_code(self):
        """Get the status code received in response.

        Returns:
            :obj:`int`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def method(self):
        """Get the method used in request.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def request_body_str(self):
        """Get the full request body.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def request_body_obj(self):
        """Get :attr:`request_body_str` deserialized into a python object.

        Returns:
            :obj:`object`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def request_object_obj(self):
        """Get the request objects from :attr:`request_body_obj`.

        Returns:
            :obj:`dict`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def response_body_str(self):
        """Get the full response body.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def response_body_obj(self):
        """Get :attr:`response_body_str` deserialized into a python object.

        Returns:
            :obj:`object`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def data_obj(self):
        """Get the response data from :attr:`response_body_obj`.

        Returns:
            :obj:`object`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def data_api(self, **kwargs):
        """Get :attr:`data_obj` deserialized into an ApiModel object.

        Returns:
            :obj:`tantrum.api_models.ApiModel`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def object_obj(self):
        """Get the response objects from :attr:`response_body_obj`.

        Returns:
            :obj:`dict`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def object_api(self, **kwargs):
        """Get :attr:`object_obj` deserialized into an ApiModel object.

        Returns:
            :obj:`tantrum.api_models.ApiModel`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def str_to_obj(self, text, src, **kwargs):
        """Deserialize string into a python object.

        Args:
            text (:obj:`str`):
                String to decode into python object
            src (:obj:`str`):
                Where text came from, used in error text.

        Raises:
            :exc:`exceptions.TextDeserializeError`:

        Returns:
            :obj:`object`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def obj_to_api(self, obj, src):
        """Deserialize a python object into :obj:`tantrum.api_models.ApiModel`.

        Args:
            obj (:obj:`object`):
                Python object to deserialize into a tantrum API object.
            src (:obj:`str`):
                Where obj came from, used in error text.

        Returns:
            :obj:`tantrum.api_models.ApiModel`

        """
        raise NotImplementedError  # pragma: no cover


class Soap(Result):
    """Deserialize SOAP API responses from Tanium API."""

    DATA_ROUTES = ["GetResultInfo", "GetResultData", "GetMergedResultData"]

    def __init__(
        self,
        api_objects,
        response_body,
        request_body,
        method,
        url,
        status_code,
        origin=None,
        lvl="info",
    ):
        """Constructor.

        Args:
            api_objects (:obj:`tantrum.api_objects.ApiObjects`):
                API Objects Container to use.
            response_body (:obj:`str`):
                Response body received from Tanium API.
            request_body (:obj:`str`):
                Request body sent to Tanium API.
            method (:obj:`str`):
                HTTP method used in request to Tanium API.
            url (:obj:`str`):
                URL used in request to Tanium API.
            status_code (:obj:`int`):
                Status code in response from Tanium API.
            origin (:obj:`requests.Response`, optional):
                Original response object received from Tanium API request.

                Defaults to: None.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        """
        self.log = utils.logs.get_obj_log(obj=self, lvl=lvl)
        """:obj:`logging.Logger`: Log for this object."""
        self._api_objects = api_objects
        self._response_body_str = response_body
        self._request_body_str = request_body
        self._method = method
        self._url = url
        self._status_code = status_code
        self._cache = {}
        self.origin = origin
        """:obj:`requests.Response`: Original response object."""

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = [
            "api_objects={!r}".format(self.api_objects),
            "url={!r}".format(self.url),
            "method={!r}".format(self.method),
            "code={!r}".format(self.status_code),
        ]
        bits = "(\n  {},\n)".format(",\n  ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @classmethod
    def from_response(cls, api_objects, response, lvl="info"):
        """Create Result from a requests response object.

        Args:
            api_objects (:obj:`tantrum.api_objects.ApiObjects`):
                API Objects Container to use.
            response (:obj:`requests.Response`):
                Response object received from Tanium API request.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        Returns:
            :obj:`Result`

        """
        return cls(
            api_objects=api_objects,
            response_body=response.text,
            request_body=response.request.body,
            method=response.request.method,
            url=response.url,
            status_code=response.status_code,
            origin=response,
            lvl=lvl,
        )

    @property
    def api_objects(self):
        """Get the API objects container.

        Returns:
            :obj:`tantrum.api_objects.ApiObjects`

        """
        return self._api_objects

    @property
    def url(self):
        """Get the URL used in request.

        Returns:
            :obj:`str`

        """
        return self._url

    @property
    def status_code(self):
        """Get the status code received in response.

        Returns:
            :obj:`int`

        """
        return int(self._status_code)

    @property
    def method(self):
        """Get the method used in request.

        Returns:
            :obj:`str`

        """
        return self._method

    @property
    def request_body_str(self):
        """Get the full request body.

        Returns:
            :obj:`str`

        """
        return self._request_body_str or ""

    @property
    def response_body_str(self):
        """Get the full response body.

        Returns:
            :obj:`str`

        """
        return self._response_body_str or ""

    def obj_to_api(self, api_name, obj, src):
        """Deserialize a python object into :obj:`tantrum.api_models.ApiModel`.

        Args:
            api_name (:obj:`str`):
                API name of obj.
            obj (:obj:`object`):
                Python object to deserialize into a tantrum API object.
            src (:obj:`str`):
                Where obj came from, used in error text.

        Returns:
            :obj:`tantrum.api_models.ApiModel`

        """
        if obj is None:
            return obj

        cls = self.api_objects.cls_by_name(name=api_name)
        with utils.tools.Timer() as t:
            if cls.__name__ == "ClientCount":
                ret = cls(count=obj)
            elif isinstance(obj, (list, tuple)):
                ret = cls(*obj)
            elif isinstance(obj, dict):
                ret = cls(**obj)
            else:
                error = "Error when deserializing API object name {a} into {c}: {o}"
                error = error.format(a=api_name, o=obj, c=cls)
                raise exceptions.ResponseError(result=self, error=error)

            # LATER(!) raise exc
        m = "Deserialized API name {a} obj type {t} from API into {cls!r}, took {e}"
        m = m.format(a=api_name, t=type(obj), cls=ret.__class__, e=t.elapsed)
        self.log.debug(m)
        return ret

    def get_dict_path(self, obj, path, src):
        """Traverse a dict using a / seperated string.

        Args:
            obj (:obj:`dict`):
                Dictionary to traverse using path.
            path (:obj:`str`):
                Nested dictionary keys seperated by / to traverse in obj.
            src (:obj:`str`):
                Where obj came from, used in error text.

        Raises:
            :exc:`exceptions.DictionaryPathError`:
                If any error occurs while traversing path in obj.

        Returns:
            :obj:`object`

        """
        try:
            ret = utils.tools.get_dict_path(obj=obj, path=path)
        except Exception as exc:
            raise exceptions.DictionaryPathError(
                result=self, obj=obj, path=path, src=src, exc=exc
            )
        return ret

    def __call__(self, raw=False, **kwargs):
        """Get :attr:`data_api` or :attr:`object_api`.

        Args:
            raw (:obj:`bool`, optional):
                Return the serialized object instead of the object deserialized into
                an ApiModel.

                Defaults to: False.
            **kwargs:
                rest of kwargs:
                    passed to :meth:`object_api` or :meth:`data_api`.

        Notes:
            Will return :meth:`object_api` or :meth:`data_api` by
            determining if :attr:`command_request` is an object or data request.

        Returns:
            :obj:`tantrum.api_models.ApiModel` or :obj:`object`

        """
        self.error_check()
        is_result_data = self.command_request in self.DATA_ROUTES
        if is_result_data:
            _raw = self.data_obj
            _api = self.data_api
        else:
            _raw = self.object_obj
            _api = self.object_api
        return _raw if raw else _api(**kwargs)

    @property
    def error_map(self):
        """Map of status codes to regex patterns and exceptions.

        Notes:
            Keys should be :obj:`int` of a status code to check patterns of.

            Values should be :obj:`list` of [pattern, exc], where pattern is a regex
            pattern to search response body, and exc is exception to throw if pattern
            match is found.

        Raises:
            :exc:`exceptions.ObjectExistsError`
            :exc:`exceptions.ObjectNotFoundError`
            :exc:`exceptions.ResponseError`

        Returns:
            :obj:`dict`

        """
        ret = {}
        ret[200] = [
            [r"400.*already.*exists", exceptions.ObjectExistsError],
            [r"NotUnique", exceptions.ObjectExistsError],
            [r"400.*not.*found", exceptions.ObjectNotFoundError],
            [r"404.*not.*found", exceptions.ObjectNotFoundError],
            [r"RBACInsufficientPrivilege", exceptions.InsufficientRBACPrivileges],
            [r"403.*", exceptions.InsufficientPrivileges],
        ]
        return ret

    def error_check(self):
        """Check for errors in :attr:`Soap.response_body_str`.

        Raises:
            :exc:`exceptions.ObjectExistsError`
            :exc:`exceptions.ObjectNotFoundError`
            :exc:`exceptions.ResponseError`

        """
        self.error_check_response()
        self.error_check_code()

    def error_check_response(self):
        """Check :attr:`Soap.command_response` for errors.

        Notes:
            For each key value pair in :attr:`error_map`, if key matches
            :attr:`Soap.status_code` or key is None, the patterns in value are
            checked against response body, and if matched the patterns associated
            exception is thrown.

        """
        for code, checks in self.error_map.items():
            if code is None or self.status_code == code:
                for check in checks:
                    pattern, error_exc = check
                    pattern_re = re.compile(pattern, re.IGNORECASE | re.DOTALL)
                    found = pattern_re.search(self.command_response)
                    if found:
                        raise error_exc(result=self, error=self.error_text)

        if self.command_request != self.command_response:
            error = "Request {req!r} does not match response:\n{resp}"
            error = error.format(resp=self.error_text, req=self.command_request)
            raise exceptions.ResponseError(result=self, error=error)

    def error_check_code(self):
        """Check if :attr:`Soap.status_code` is one of :attr:`valid_codes`.

        Raises:
            :exc:`exceptions.ResponseError`

        """
        if self.status_code not in self.valid_codes:
            error = "Response status code {c} is not one of {v}, error text:\n{e}"
            error = error.format(
                c=self.status_code, v=self.valid_codes, e=self.error_text
            )
            raise exceptions.ResponseError(result=self, error=error)

    @property
    def valid_codes(self):
        """List of valid status codes.

        Returns:
            :obj:`list` of :obj:`int`

        """
        return [200]

    @property
    def error_text(self):
        """Get the error text from :attr:`Soap.response_body_str`.

        Returns:
            :obj:`str`

        """
        return ", ".join([x for x in self.command_response.splitlines() if x])

    @property
    def command_request(self):
        """Get the "command" element from :attr:`request_body_obj`.

        Returns:
            :obj:`str`

        """
        obj = self.request_body_obj
        path = "soap:Envelope/soap:Body/t:tanium_soap_request/command"
        src = "SOAP API deserialized request body"
        return self.get_dict_path(obj=obj, path=path, src=src)

    @property
    def command_response(self):
        """Get the "command" element from :attr:`response_body_obj`.

        Returns:
            :obj:`str`

        """
        obj = self.response_body_obj
        path = "soap:Envelope/soap:Body/t:return/command"
        src = "SOAP API deserialized response body"
        return self.get_dict_path(obj=obj, path=path, src=src)

    @property
    def request_body_obj(self):
        """Deserialize :attr:`Soap.request_body_str` into a python object.

        Returns:
            :obj:`dict`

        """
        key = "request_body_obj"
        if key not in self._cache:
            text = self.request_body_str
            src = "SOAP API request body"
            self._cache[key] = self.str_to_obj(
                text=text, src=src, try_int=True, use_dict=True, flat_attrs=True
            )
        return self._cache[key]

    @property
    def request_object_obj(self):
        """Get the request objects from :attr:`request_body_obj`.

        Returns:
            :obj:`dict`

        """
        obj = self.request_body_obj
        path = "soap:Envelope/soap:Body/t:tanium_soap_request/object_list"
        src = "SOAP API deserialized request body"
        return self.get_dict_path(obj=obj, path=path, src=src)

    @property
    def response_body_obj(self):
        """Get :attr:`Soap.response_body_str` deserialized into a python object.

        Returns:
            :obj:`dict`

        """
        key = "response_body_obj"
        if key not in self._cache:
            text = self.response_body_str
            src = "SOAP API response body"
            self._cache[key] = self.str_to_obj(
                text=text, src=src, try_int=True, use_dict=True, flat_attrs=True
            )
        return self._cache[key]

    @property
    def data_xml(self):
        """Get the "ResultXML" element from :attr:`response_body_obj`.

        Returns:
            :obj:`str`

        """
        obj = self.response_body_obj
        path = "soap:Envelope/soap:Body/t:return/ResultXML"
        src = "SOAP API deserialized response body"
        data = self.get_dict_path(obj=obj, path=path, src=src)
        return data or ""

    @property
    def data_obj(self):
        """Get the response data from :attr:`data_xml`.

        Returns:
            :obj:`object`

        """
        key = "data_obj"
        if key not in self._cache:
            text = self.data_xml
            src = "'ResultXML' element from SOAP API deserialized response body"
            self._cache[key] = self.str_to_obj(
                text=text, src=src, try_int=False, use_dict=True, flat_attrs=True
            )
        return self._cache[key]

    def data_api(self, **kwargs):
        """Get :attr:`data_obj` deserialized into an ApiModel object.

        Args:
            **kwargs:
                rest of kwargs:
                    passed to :meth:`Soap.obj_to_api`

        Returns:
            :obj:`tantrum.api_models.ApiModel`

        """
        if self.command_request not in self.DATA_ROUTES:
            error = "Route {req!r} is not one of {data}, not a data request?"
            error = error.format(req=self.command_request, data=self.DATA_ROUTES)
            raise exceptions.ApiWrongRequestType(result=self, error=error)

        kwargs["src"] = "Result data from 'ResultXML' element in SOAP response"
        kwargs["api_name"] = list(self.data_obj.keys())[0]
        kwargs["obj"] = list(self.data_obj.values())[0]
        return self.obj_to_api(**kwargs)

    @property
    def object_obj(self):
        """Get the response objects from :attr:`response_body_obj`.

        Returns:
            :obj:`dict`

        """
        obj = self.response_body_obj
        path = "soap:Envelope/soap:Body/t:return/result_object"
        src = "'result_object' element from SOAP API deserialized response body"
        data = self.get_dict_path(obj=obj, path=path, src=src)
        return data or {}

    def object_api(self, **kwargs):
        """Get :attr:`object_obj` deserialized into an ApiModel object.

        Args:
            **kwargs:
                rest of kwargs:
                    passed to :meth:`Soap.obj_to_api`

        Returns:
            :obj:`tantrum.api_models.ApiModel`

        """
        if self.command_request in self.DATA_ROUTES:
            error = "Route {route!r} is not one of {data}, not a data request?"
            error = error.format(route=self.command_request, data=self.DATA_ROUTES)
            raise exceptions.ApiWrongRequestType(result=self, error=error)
        kwargs["src"] = "Result object from 'result_object' element in SOAP response"
        kwargs["api_name"] = list(self.object_obj.keys())[0]
        kwargs["obj"] = list(self.object_obj.values())[0]
        return self.obj_to_api(**kwargs)

    def str_to_obj(
        self, text, src, use_dict=True, try_int=True, flat_attrs=True, **kwargs
    ):
        """Deserialize string into a python object.

        Args:
            text (:obj:`str`):
                String to decode into python object
            src (:obj:`str`):
                Where text came from, used in error text.
            use_dict (:obj:`bool`):
                Use :obj:`dict` instead of :obj:`collections.OrderedDict` when
                deserializing.

                Defaults to: True.
            try_int (:obj:`bool`):
                Try to convert str into int when deserializing.

                Defaults to: True.
            flat_attrs (:obj:`bool`):
                Make attr_prefix "" and cdata_key "text".

                Defaults to: True.
            **kwargs:
                rest of kwargs:
                    Passed to xmltodict.parse.

        Raises:
            :exc:`exceptions.TextDeserializeError`:
                If an exception is thrown from xmltodict.parse.

        Returns:
            :obj:`dict`

        """
        if use_dict:
            kwargs.setdefault("dict_constructor", dict)
        if try_int:
            kwargs.setdefault("postprocessor", try_int_xml)
        if flat_attrs:
            kwargs.setdefault("attr_prefix", "")
            kwargs.setdefault("cdata_key", "text")

        with utils.tools.Timer() as t:
            if text:
                try:
                    ret = xmltodict.parse(text, **kwargs)
                except Exception as exc:
                    raise exceptions.TextDeserializeError(
                        result=self, text=text, src=src, exc=exc
                    )
            else:
                ret = {}
        m = "Finished deserializing {src} into {t}, {size} took {e}"
        m = m.format(size=len(text), src=src, t=type(ret), e=t.elapsed)
        self.log.debug(m)
        return ret

    @staticmethod
    def pretty_xml(obj, **kwargs):
        return pretty_xml(obj=obj, **kwargs)

    def pretty_bodies(self, as_str=True):
        return pretty_bodies(
            response_body=self.response_body_str,
            request_body=self.request_body_str,
            as_str=as_str,
        )


def try_int_xml(path, key, value):
    """Parser hook for xmltodict.parse to try to convert str to int.

    Args:
        path (:obj:`list`):
            Path from root of document to key.
        key (:obj:`str`):
            Element name being parsed.
        value (:obj:`object`):
            Element value of key being parsed.

    Returns:
        (:obj:`str`, :obj:`object`):
            original key supplied and value as int if successfully converted.

    """
    if isinstance(value, six.string_types):
        try:
            return key, int(value)
        except (ValueError, TypeError):
            return key, value
    return key, value


def pretty_xml(obj, **kwargs):
    """Re-parse an XML string into a pretty XML string.

    Args:
        obj (:obj:`object`):
            Python object to encode into a string.
        **kwargs:
            pretty (:obj:`bool`):
                Indent the output doc.

                Defaults to: True.
            rest of kwargs:
                Passed to xmltodict.unparse.

    Returns:
        :obj:`str`

    """
    kwargs.setdefault("pretty", True)
    return xmltodict.unparse(xmltodict.parse(obj), **kwargs)


def pretty_bodies(response_body, request_body, as_str=True):
    """Return pretty XML of last response and request body from http_client.

    Args:
        http_client (:obj:`tantrum.http_client.HttpClient`):
            HTTP Client to get last_response attribute from
        as_str (:obj:`bool`, optional):
            Return pretty XML as a string instead of a dict.

            Defaults to: True.

    Returns:
        :obj:`dict` or :obj:`str`

    """
    ret = {"response": response_body, "request": request_body}
    try:
        ret["response"] = pretty_xml(response_body)
    except Exception as exc:
        ret["response_exc"] = exc
    try:
        ret["request"] = pretty_xml(request_body)
    except Exception as exc:
        ret["request_exc"] = exc
    if as_str:
        ret_list = [
            "",
            "<!-- Request Body -->",
            "",
            ret["request"],
            "",
            "<!-- Response Body -->",
            "",
            ret["response"],
        ]
        ret = "\n".join(ret_list)
    return ret



# -*- coding: utf-8 -*-
"""tantrum versioning utilities module."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six

from . import exceptions


def version_recomp(v1, v2, vshrink=True):
    """Recompile version strings into lists.

    Args:
        v1 (:obj:`str` or :obj:`list`):
            Version that is being compared against v2.

            Will be shrunk down to the same dotted length as v2 if vshrink is True.

            Will be split into list using period.
        v2 (:obj:`str` or :obj:`list`):
            Version that v1 is being compared against.

            Will be split into list using period.
        vshrink (:obj:`bool` or :obj:`int`, optional):
            If True, shrink v1 down to match the same length as v2.

            If False, do not shrink v1 at all.

            If int, shrink both v1 and v2 down to this length.

            Defaults to: True.

    Notes:
        If any non-integers found in v1 or v2, they will be returned as joined
        strings using periods.

        Default python string comparison leads to a less-than-perfect comparison
        for ints mixed with strs.

    Returns:
        (:obj:`list` of :obj:`str`, :obj:`list` of :obj:`str`) or
        (:obj:`str`, :obj:`str`)

    """
    v1, v2 = split_ver(v1), split_ver(v2)
    v2_len = len(v2)
    if isinstance(vshrink, bool):
        if vshrink:
            v1 = v1[:v2_len]
    elif isinstance(vshrink, int):
        v1 = v1[:vshrink]
        v2 = v2[:vshrink]
    if any(not isinstance(x, int) for x in v1 + v2):
        v1, v2 = join_ver(v1), join_ver(v2)
    return v1, v2


def version_eq(v1, v2, vshrink=True):
    """Check if v1 is equal to v2.

    Args:
        v1 (:obj:`str` or :obj:`list`):
            Version that is being compared against v2.

            Will be shrunk down to the same dotted length as v2 if vshrink is True.
        v2 (:obj:`str` or :obj:`list`):
            Version that v1 is being compared against.
        vshrink (:obj:`bool` or :obj:`int`, optional):
            If True, shrink v1 down to match the same length as v2.

            If False, do not shrink v1 at all.

            If int, shrink both v1 and v2 down to this length.

            Defaults to: True.

    Returns:
        :obj:`bool`

    """
    v1, v2 = version_recomp(v1=v1, v2=v2, vshrink=vshrink)
    return v1 == v2


def version_min(v1, v2, vshrink=True):
    """Check if v1 is greater than or equal to v2.

    Args:
        v1 (:obj:`str` or :obj:`list`):
            Version that is being compared against v2.

            Will be shrunk down to the same dotted length as v2 if vshrink is True.
        v2 (:obj:`str` or :obj:`list`):
            Version that v1 is being compared against.
        vshrink (:obj:`bool` or :obj:`int`, optional):
            If True, shrink v1 down to match the same length as v2.

            If False, do not shrink v1 at all.

            If int, shrink both v1 and v2 down to this length.

            Defaults to: True.

    Returns:
        :obj:`bool`

    """
    v1, v2 = version_recomp(v1=v1, v2=v2, vshrink=vshrink)
    return v1 >= v2


def version_max(v1, v2, vshrink=True):
    """Check if v1 is less than or equal to v2.

    Args:
        v1 (:obj:`str` or :obj:`list`):
            Version that is being compared against veq, vmax, and vmin.

            Will be be shrunk down to the same dotted length as veq, vmax, and vmin
            if vshrink is True.
        v2 (:obj:`str` or :obj:`list`):
            Version that v1 is being compared against.
        vshrink (:obj:`bool` or :obj:`int`, optional):
            If True, shrink v1 down to match the same length as v2.

            If False, do not shrink v1 at all.

            If int, shrink both v1 and v2 down to this length.

            Defaults to: True.

    Returns:
        :obj:`bool`

    """
    v1, v2 = version_recomp(v1=v1, v2=v2, vshrink=vshrink)
    return v1 <= v2


def version_check(version, veq="", vmax="", vmin="", vshrink=True):
    """Check if version meets veq, vmax, and vmin.

    Args:
        version (:obj:`str` or :obj:`list`):
            Version that is being compared against veq, vmax, and vmin.

            Will be be shrunk down to the same dotted length as veq, vmax, and vmin
            if vshrink is True.
        veq (:obj:`str`, optional):
            Exact version that version must match.

            No comparison performed if empty.

            Defaults to: "".
        vmin (:obj:`str`, optional):
            Minimum version that version must match.

            No comparison performed if empty.

            Defaults to: "".
        vmax (:obj:`str`, optional):
            Maximum version that version must match.

            No comparison performed if empty.

            Defaults to: "".
        vshrink (:obj:`bool` or :obj:`int`, optional):
            If True, shrink version down to match the same length as
            veq, vmax, and vmin.

            If False, do not shrink version at all.

            If int, shrink version, veq, vmax, and vmin down to this length.

            Defaults to: True.

    Returns:
        :obj:`bool`

    """
    if veq and not version_eq(v1=version, v2=veq, vshrink=vshrink):
        return False
    if vmax and not version_max(v1=version, v2=vmax, vshrink=vshrink):
        return False
    if vmin and not version_min(v1=version, v2=vmin, vshrink=vshrink):
        return False
    return True


def version_check_obj_req(version, src, obj):
    """Check if veq, vmax, and vmin from a classes get_version_req meets version.

    Args:
        version (:obj:`str` or :obj:`list`):
            Version that is being compared against veq, vmax, and vmin keys
            in the dict returned from the get_version_req class of obj.

            Will be be shrunk down to the same dotted length as veq, vmax, and vmin
            if vshrink is True.

        src (:obj:`str`):
            Where version came from, used in exception message.
        obj (:obj:`object`):
            Object or class that has get_version_req class method.

    Raises:
        :exc:`tantrum.utils.exceptions.VersionMismatchError`:
              If any of veq, vmax, vmin keys in dict returned by get_version_req
              do not match version.

    """
    version_req = obj.get_version_req()
    vmin = version_req.get("vmin", "")
    vmax = version_req.get("vmax", "")
    veq = version_req.get("veq", "")
    if not version_check(version=version, veq=veq, vmax=vmax, vmin=vmin):
        reqs = ", ".join(
            [
                "a minimum of {!r}".format(vmin or None),
                "a maximum of {!r}".format(vmax or None),
                "and equal to {!r}".format(veq or None),
            ]
        )
        error = "\n  ".join(
            [
                "",
                "{o} failed a version requirement!",
                "Version must be {reqs}",
                "{src!r} is version {v!r}",
            ]
        )
        error = error.format(o=obj, reqs=reqs, v=version, src=src)
        raise exceptions.VersionMismatchError(error)


def split_ver(v):
    """Split v into a list of integers.

    Args:
        v (:obj:`str` or :obj:`list` of :obj:`int` or :obj:`int`):
            Version string to split using periods.

    Returns:
        :obj:`list` of :obj:`int` or :obj:`str`

    """
    v = v.split(".") if isinstance(v, six.string_types) else v
    return [
        int(x) if isinstance(x, six.string_types) and x.isdigit() else x
        for x in v
        if x not in ["", None]
    ]


def join_ver(v):
    """Join v into a dot string.

    Args:
        v (:obj:`str` or :obj:`list` of :obj:`int` or :obj:`int`):
            Version string to join using periods.

    Returns:
        :obj:`str`

    """
    return ".".join([format(x) for x in v]) if isinstance(v, list) else v



# -*- coding: utf-8 -*-
"""Workflow encapsulation package for performing actions using the Tanium API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import json
import time

from collections import OrderedDict

from . import exceptions
from .. import utils
from .. import results


class Workflow(object):
    def __init__(self, adapter, obj, lvl="info", result=None):
        """Constructor.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for this workflow.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".
            result (:obj:`tantrum.results.Result`, optional):
                Result object that ``obj`` was generated from.

                Defaults to: None.

        """
        self._lvl = lvl
        self.log = utils.logs.get_obj_log(obj=self, lvl=lvl)
        self.obj = obj
        self.adapter = adapter
        self._result = result
        self._last_result = result

    def __repr__(self):
        """Show object info.

        Returns:
            (:obj:`str`)

        """
        return self.__str__()

    @property
    def api_objects(self):
        return self.adapter.api_objects


class Clients(Workflow):
    def __str__(self):
        """Show object info.

        Returns:
            (:obj:`str`)

        """
        ctmpl = "{c.__module__}.{c.__name__}".format
        bits = ["count={}".format(len(self.obj))]
        bits = "({})".format(", ".join(bits))
        cls = ctmpl(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    @staticmethod
    def build_last_reg_filter(
        adapter, last_reg=300, operator="greaterequal", not_flag=False, filters=None
    ):
        """Build a set of filters to be used in :meth:`Clients.get_all.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            last_reg (:obj:`int`, optional):
                Only return clients that have registered in N number of seconds.

                Defaults to: 300.
            operator (:obj:`str`, optional):
                Defines how the last_registered attribute of each client status is
                compared against the value in last_reg.
                Must be one of :data:`OPERATOR_MAP`.

                Defaults to: "greaterequal".
            not_flag (:obj:`int`, optional):
                If True have the API return all rows that do not match the operator.

                Defaults to: 1000.
            filters (:obj:`object`, optional):
                If a CacheFilterList object is supplied, the last_registration filter
                generated by this method will be appended to it. If this is None,
                a new CacheFilterList will be created with the last_registration filter
                being the only item in it.

                Defaults to: None.

        Returns:
            :obj:`Clients`

        """
        op_dict = get_operator_map(operator)
        now_dt = datetime.datetime.utcnow()
        ago_td = datetime.timedelta(seconds=-(int(last_reg)))
        ago_dt = now_dt + ago_td
        ago_str = ago_dt.strftime(adapter.api_objects.module_dt)
        cfilter = adapter.api_objects.CacheFilter(
            field="last_registration",
            type="Date",
            operator=op_dict["op"],
            not_flag=not_flag,
            value=ago_str,
        )
        filters = filters or adapter.api_objects.CacheFilterList()
        filters.append(cfilter)
        return filters

    @classmethod
    def get_all(
        cls,
        adapter,
        filters=None,
        sort_fields="last_registration",
        page_size=1000,
        max_page_count=0,
        cache_expiration=600,
        sleep=5,
        lvl="info",
        **kwargs
    ):
        """Get all Clients.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            filters (:obj:`object`, optional):
                Tantrum CacheFilterList returned from
                :meth:`Clients.build_last_reg_filter`.

                Defaults to: None.
            sort_fields (:obj:`str`, optional):
                Attribute of a ClientStatus object to have API sort the return on.

                Defaults to: "last_registration".
            page_size (:obj:`int`, optional):
                Get N number of clients at a time from the API.
                If 0, disables paging and gets all clients in one call.

                Defaults to: 1000.
            max_page_count (:obj:`int`, optional):
                Only fetch up to this many pages. If 0, get all pages.

                Defaults to: 0.
            cache_expiration (:obj:`int`, optional):
                When page_size is not 0, have the API keep the cache of clients
                for this many seconds before expiring the cache.

                Defaults to: 600.
            sleep (:obj:`int`, optional):
                Wait N seconds between fetching each page.

                Defaults to: 5.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get`.

        Returns:
            :obj:`Clients`

        """
        log = utils.logs.get_obj_log(obj=cls, lvl=lvl)

        get_args = {}
        get_args.update(kwargs)
        get_args["cache_sort_fields"] = sort_fields
        get_args["obj"] = adapter.api_objects.ClientStatus()

        row_start = 0
        row_count = page_size

        if filters is not None:
            get_args["cache_filters"] = filters

        if page_size:
            get_args["row_start"] = row_start
            get_args["row_count"] = row_count
            get_args["cache_expiration"] = cache_expiration

        result = adapter.cmd_get(**get_args)

        log.debug(result.pretty_bodies())

        result_obj = result()

        m = "Received initial {o!r} length={len}, cache_info={cache!r}"
        m = m.format(
            o=result_obj.__class__.__name__,
            len=len(result_obj),
            cache=getattr(result_obj, "cache_info", None),
        )
        log.info(m)

        if page_size:
            total_rows = result_obj.cache_info.filtered_row_count
            paging_get_args = {k: v for k, v in get_args.items()}
            page_count = 1

            while True:

                if max_page_count and page_count >= max_page_count:
                    m = "Reached max page count {c}, considering all clients fetched"
                    m = m.format(c=max_page_count)
                    log.info(m)
                    break

                if len(result_obj) >= total_rows:
                    m = "Reached total rows count {c}, considering all clients fetched"
                    m = m.format(c=total_rows)
                    log.info(m)
                    break

                page_count += 1
                row_start += row_count

                paging_get_args["row_start"] = row_start

                result = adapter.cmd_get(**paging_get_args)

                log.debug(result.pretty_bodies())

                paging_result_obj = result()

                m = "Received page of {o!r} length={len}, cache_info={cache!r}"
                m = m.format(
                    o=paging_result_obj.__class__.__name__,
                    len=len(paging_result_obj),
                    cache=getattr(paging_result_obj, "cache_info", None),
                )
                log.info(m)

                result_obj += paging_result_obj

                m = "{o!r} received so far {len} out of total {total}"
                m = m.format(
                    o=result_obj.__class__.__name__,
                    len=len(result_obj),
                    total=total_rows,
                )
                log.info(m)

                time.sleep(sleep)

            return cls(adapter=adapter, obj=result_obj, lvl=lvl, result=result)


class Sensor(Workflow):
    def __str__(self):
        """Show object info.

        Returns:
            (:obj:`str`)

        """
        ctmpl = "{c.__module__}.{c.__name__}".format
        bits = [
            "name={!r}".format(self.obj.name),
            "filter={}".format(", ".join(self.filter_vals)),
        ]
        if self.params_defined or self.param_values:
            bits += [
                "params_defined={}".format(list(self.params_defined.keys())),
                "param_values={}".format(list(self.param_values.items())),
            ]
        bits = "({})".format(", ".join(bits))
        cls = ctmpl(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    @classmethod
    def get_by_name(cls, adapter, name, lvl="info"):
        """Get a sensor object by name.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            name (:obj:`str`):
                Name of sensor to fetch.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        Returns:
            :obj:`Sensor`

        """
        result = adapter.cmd_get(obj=adapter.api_objects.Sensor(name=name))
        return cls(adapter=adapter, obj=result(), lvl=lvl, result=result)

    @classmethod
    def get_by_id(cls, adapter, id, lvl="info"):
        """Get a sensor object by id.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            id (:obj:`int`):
                id of sensor to fetch.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        Returns:
            :obj:`Sensor`

        """
        result = adapter.cmd_get(obj=adapter.api_objects.Sensor(id=id))
        return cls(adapter=adapter, obj=result(), lvl=lvl, result=result)

    @property
    def params_defined(self):
        """Get the parameter definitions for this sensor.

        Notes:
            Will try to resolve a default value and store it in "derived_default" key
            for each parameter definition returned.

        Returns:
            :obj:`collections.OrderedDict`

        """
        param_defs = json.loads(self.obj.parameter_definition or "{}")
        params = param_defs.get("parameters", [])
        for p in params:
            pdef = p.get("defaultValue", "")
            pval = p.get("value", "")
            pvals = p.get("values", [])
            if pdef not in ["", None]:
                derived_default = pdef
            elif pval not in ["", None]:
                derived_default = pval
            elif pvals:
                derived_default = pvals[0]
            else:
                derived_default = ""
            p["derived_default"] = derived_default
        return OrderedDict((p["key"], p) for p in params)

    @property
    def param_values(self):
        """Get all of the parameter key and values.

        Returns:
            :obj:`OrderedDict`

        """
        ret = OrderedDict()
        for k in self.params_defined:
            ret[k] = ""
        for p in self.params:
            ret[p.key] = p.value
        return ret

    @property
    def params(self):
        """Get the parameters that are set for this sensor.

        Returns:
            :obj:`tantrum.api_objects.ApiObjects`: ParameterList API object

        """
        if not hasattr(self, "_params"):
            self._params = self.api_objects.ParameterList()
        return self._params

    def set_parameter(
        self, key, value="", derive_default=True, delim="||", allow_undefined=True
    ):
        """Set a parameters value for this sensor.

        Args:
            key (:obj:`str`):
                Key name of parameter to set.
            value (:obj:`str`, optional):
                Value of parameter to set.

                Defaults to: "".
            derive_default (:obj:`bool`, optional):
                Get default value from parameter definition if value is "".

                Defaults to: True.
            delim (:obj:`str`, optional):
                String to put before and after parameter key name when sending to API.

                Defaults to: "||".
            allow_undefined (:obj:`bool`, optional):
                Allow parameter keys that are not in the parameters definition
                for this sensor to be set.
                Throws exception if False and key not in :attr:`Sensor.param_keys`.

                Defaults to: True.

        """
        param_def = self.params_defined.get(key, None)
        if param_def is None:
            m = "Parameter key {o!r} is not one of the defined parameters {ov}"
            m = m.format(o=key, ov=list(self.params_defined.keys()))
            if allow_undefined:
                self.log.info(m)
            else:
                raise exceptions.ModuleError(m)
        elif derive_default and value == "":
            value = param_def.get("derived_default", "")

        key_delim = "{d}{key}{d}".format(d=delim, key=key)
        param = self.api_objects.Parameter(key=key_delim, value=value)

        self.params.append(param)

    @property
    def filter(self):
        """Get the filter for this sensor.

        Returns:
            :obj:`tantrum.api_objects.ApiObjects`: Filter API object

        """
        if not hasattr(self, "_filter"):
            self._filter = self.api_objects.Filter()
            self._filter.sensor = self.api_objects.Sensor()
            self._filter.sensor.hash = self.obj.hash
        return self._filter

    @property
    def filter_vals(self):
        """Get the key value pairs of the filter for this sensor.

        Returns:
            :obj:`list` of :obj:`str`

        """
        if any([self.filter.value, self.filter.operator]):
            keys = [
                "operator",
                "value",
                "ignore_case_flag",
                "not_flag",
                "all_values_flag",
                "max_age_seconds",
                "value_type",
            ]
            vals = ["{}: {!r}".format(k, getattr(self.filter, k)) for k in keys]
        else:
            vals = []
        return vals

    def set_filter(
        self,
        value,
        operator="regex",
        ignore_case_flag=True,
        not_flag=False,
        all_values_flag=False,
        max_age_seconds=0,
        type=None,
    ):
        """Set a filter for this sensor to be used in a question.

        Args:
            value (:obj:`str`):
                Filter sensor rows returned on this value.
            operator (:obj:`str`, optional):
                Operator to use for filter_value.
                Must be one of :data:`OPERATOR_MAP`.

                Defaults to: "regex".
            ignore_case_flag (:obj:`bool`, optional):
                Ignore case when filtering on value.

                Defaults to: True.
            not_flag (:obj:`bool`, optional):
                If set, negate the match.

                Defaults to: False.
            max_age_seconds (:obj:`int`, optional):
                How old a sensor result can be before we consider it invalid.
                0 means to use the max age property of the sensor.

                Defaults to: 0.
            all_values_flag (:obj:`bool`, optional):
                Have filter match all values instead of any value.

                Defaults to: False.
            type (:obj:`str`, optional):
                Have filter consider the value type as this.
                Must be one of :data:`TYPE_MAP`

                Defaults to: None.

        """
        op_dict = get_operator_map(operator)
        if type:
            get_type_map(type)

        self.filter.value = op_dict["tmpl"].format(value=value)
        self.filter.operator = op_dict["op"]
        self.filter.ignore_case_flag = ignore_case_flag
        self.filter.not_flag = not_flag
        self.filter.all_values_flag = all_values_flag
        self.filter.max_age_seconds = max_age_seconds
        self.filter.value_type = type

    def build_select(self, set_param_defaults=True, allow_empty_params=False):
        select = self.api_objects.Select()
        select.filter = self.filter
        select.sensor = self.api_objects.Sensor()

        for key in self.params_defined:
            if key not in self.param_values and set_param_defaults:
                self.set_parameter(key=key, derive_default=True)

        for param in self.params:
            if param.value in ["", None] and not allow_empty_params:
                m = "Parameter {p.key!r} value {p.value!r} is empty, definition: {d}"
                m = m.format(p=param, d=self.params_defined.get(key, None))
                raise exceptions.ModuleError(m)

        if self.params:
            select.sensor.parameters = self.params
            select.sensor.source_id = self.obj.id
            select.filter.sensor.id = self.obj.id
        else:
            select.sensor.hash = self.obj.hash
        select.WORKFLOW = self
        return select


class Question(Workflow):
    def __str__(self):
        """Show object info.

        Returns:
            (:obj:`str`)

        """
        ctmpl = "{c.__module__}.{c.__name__}".format
        atmpl = "{k}='{v}'".format
        attrs = ["id", "query_text"]
        bits = [atmpl(k=attr, v=getattr(self.obj, attr, None)) for attr in attrs]
        bits += [atmpl(k=k, v=v) for k, v in self.expiration.items()]
        bits = "(\n  {},\n)".format(",\n  ".join(bits))
        cls = ctmpl(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    @classmethod
    def new(cls, adapter, lvl="info"):
        """Create a new Question workflow.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        Returns:
            :obj:`Question`

        """
        return cls(obj=adapter.api_objects.Question(), adapter=adapter, lvl=lvl)

    @classmethod
    def get_by_id(cls, adapter, id, lvl="info"):
        """Get a question object by id.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            id (:obj:`int`):
                id of question to fetch.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        Returns:
            :obj:`Question`

        """
        result = adapter.cmd_get(obj=adapter.api_objects.Question(id=id))
        return cls(adapter=adapter, obj=result(), lvl=lvl, result=result)

    def _check_id(self):
        """Check that question has been asked by seeing if self.obj.id is set."""
        if not self.obj.id:
            m = "No id issued yet, ask the question!"
            raise exceptions.ModuleError(m)

    @property
    def expiration(self):
        """Get expiration details for this question.

        Returns:
            :obj:`dict`

        """
        now_dt = datetime.datetime.utcnow()
        now_td = datetime.timedelta()
        ret = {
            "expiration": now_dt,
            "expire_in": now_td,
            "expire_ago": now_td,
            "expired": True,
        }
        if self.obj.expiration:
            ex_dt = self.api_objects.module_dt_format(self.obj.expiration)
            is_ex = now_dt >= ex_dt
            ret["expiration"] = ex_dt
            ret["expired"] = is_ex
            if is_ex:
                ret["expire_ago"] = now_dt - ex_dt
            else:
                ret["expire_in"] = ex_dt - now_dt
        return ret

    def refetch(self):
        """Re-fetch this question."""
        self._check_id()
        result = self.adapter.cmd_get(obj=self.obj)
        self._last_result = result
        self.obj = result()

    def ask(self, **kwargs):
        """Ask the question.

        Args:
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_add`.

        Notes:
            If question has already been asked (id is set), we wipe out attrs:
            ["id", "context_group", "management_rights_group"], then add it.

        """
        if self.obj.id:
            wipe_attrs = ["id", "context_group", "management_rights_group"]
            for attr in wipe_attrs:
                setattr(self.obj, attr, None)
        result = self.adapter.cmd_add(obj=self.obj, **kwargs)
        self._last_result = result
        self.obj = result()
        self.refetch()

    def add_left_sensor(
        self, sensor, set_param_defaults=True, allow_empty_params=False
    ):
        """Add a sensor to the left hand side of the question.

        Args:
            sensor (:obj:`Sensor`):
                Sensor workflow object.
            set_param_defaults (:obj:`bool`, optional):
                If sensor has parameters defined, and no value is set,
                try to derive the default value from each parameters definition.

                Defaults to: True.
            allow_empty_params (:obj:`bool`, optional):
                If sensor has parameters defined, and the value is not set, "", or None,
                throw an exception.

                Defaults to: True.

        """
        select = sensor.build_select(
            set_param_defaults=set_param_defaults, allow_empty_params=allow_empty_params
        )
        if not getattr(self.obj, "selects", None):
            self.obj.selects = self.api_objects.SelectList()
        self.obj.selects.append(select)

    def answers_get_info(self, **kwargs):
        """Return the ResultInfo for this question.

        Args:
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get_result_info`.

        Returns:
            :obj:`tantrum.api_models.ApiModel`: ResultInfoList API Object

        """
        self._check_id()

        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj

        result = self.adapter.cmd_get_result_info(**cmd_args)

        self._last_result = result

        infos = result()

        self._last_infos = infos

        m = "Received answers info: {infos}"
        m = m.format(infos=infos.serialize())
        self.log.debug(m)
        self.log.debug(format(self))

        return infos

    def answers_poll(
        self,
        poll_pct=99,
        poll_secs=0,
        poll_total=0,
        poll_sleep=5,
        max_poll_count=0,
        **kwargs
    ):
        """Poll for answers from clients for this question.

        Args:
            poll_sleep (:obj:`int`, optional):
                Check for answers every N seconds.

                Defaults to: 5.
            poll_pct (:obj:`int`, optional):
                Wait until the percentage of clients total is N percent.

                Defaults to: 99.
            poll_secs (:obj:`int`, optional):
                If not 0, wait until N seconds for pct of clients total instead of
                until question expiration.

                Defaults to: 0.
            poll_total (:obj:`int`, optional):
                If not 0, wait until N clients have total instead of
                ``estimated_total`` of clients from API.

                Defaults to: 0.
            max_poll_count (:obj:`int`, optional):
                If not 0, only poll N times.

                Defaults to: 0.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`answers_get_info`.

        Returns:
            :obj:`object`: ResultInfoList API object

        """
        # TODO: Add wait till error_count / no_results_count == 0
        self._check_id()

        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj

        start = datetime.datetime.utcnow()

        if poll_secs:
            stop_dt = start + datetime.timedelta(seconds=poll_secs)
        else:
            stop_dt = self.expiration["expiration"]

        m = "Start polling loop for answers until for {o} until {stop_dt}"
        m = m.format(o=self, stop_dt=stop_dt)
        self.log.debug(m)

        infos = self.answers_get_info(**kwargs)
        info = infos[0]
        est_total = info.estimated_total
        poll_total = est_total

        if poll_total and poll_total <= est_total:
            this_total = poll_total

        now_pct = utils.tools.calc_percent(part=info.mr_passed, whole=this_total)

        poll_count = 0
        while True:
            poll_count += 1
            m = "New polling loop #{c} for {o}"
            m = m.format(c=poll_count, o=self)
            self.log.debug(m)

            if now_pct >= poll_pct:
                m = "Reached {now_pct} out of {pct}, considering all answers in"
                m = m.format(now_pct=PCT_FMT(now_pct), pct=PCT_FMT(poll_pct))
                self.log.info(m)
                break

            if datetime.datetime.utcnow() >= stop_dt:
                m = "Reached stop_dt {stop_dt}, considering all answers in"
                m = m.format(stop_dt=stop_dt)
                self.log.info(m)
                break

            if self.expiration["expired"]:
                m = "Reached expiration {expiration}, considering all answers in"
                m = m.format(expiration=self.expiration)
                self.log.info(m)
                break

            if max_poll_count and poll_count >= max_poll_count:
                m = "Reached max poll count {c}, considering all answers in"
                m = m.format(c=max_poll_count)
                self.log.info(m)
                break

            infos = self.answers_get_info(**kwargs)
            info = infos[0]

            now_pct = utils.tools.calc_percent(part=info.mr_passed, whole=this_total)

            m = [
                "Answers in {now_pct} out of {pct}",
                "{info.mr_passed} out of {this_total}",
                "estimated_total: {info.estimated_total}",
                "poll count: {c}",
            ]
            m = ", ".join(m)
            m = m.format(
                now_pct=PCT_FMT(now_pct),
                pct=PCT_FMT(poll_pct),
                info=info,
                this_total=this_total,
                c=poll_count,
            )
            self.log.info(m)

            time.sleep(poll_sleep)

        end = datetime.datetime.utcnow()
        elapsed = end - start

        m = [
            "Finished polling in: {dt}",
            "clients answered: {info.mr_passed}",
            "estimated clients: {info.estimated_total}",
            "rows in answers: {info.row_count}",
            "poll count: {c}",
        ]
        m = ", ".join(m)
        m = m.format(dt=elapsed, info=info, c=poll_count)
        self.log.info(m)

        return infos

    def answers_get_data(self, hashes=False, **kwargs):
        """Get the answers for this question.

        Args:
            hashes (:obj:`bool`, optional):
                Have the API include the hashes of rows values.

                Defaults to: False.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get_result_data`.

        Notes:
            This will not use any paging, which means ALL answers will be returned
            in one API response. For large data sets of answers, this is unwise.

        Returns:
            :obj:`tantrum.api_models.ApiModel`: ResultDataList API Object

        """
        self._check_id()

        start = datetime.datetime.utcnow()

        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj
        cmd_args["include_hashes_flag"] = hashes

        result = self.adapter.cmd_get_result_data(**cmd_args)
        self._last_result = result

        end = datetime.datetime.utcnow()
        elapsed = end - start

        m = "Finished getting answers in {dt}"
        m = m.format(dt=elapsed)
        self.log.info(m)

        datas = result()
        self._last_datas = datas
        return datas

    def answers_get_data_paged(
        self,
        page_size=1000,
        max_page_count=0,
        max_row_count=0,
        cache_expiration=900,
        hashes=False,
        sleep=5,
        **kwargs
    ):
        """Get the answers for this question one page at a time.

        Args:
            page_size (:obj:`int`, optional):
                Size of each page to fetch at a time.

                Defaults to: 1000.
            max_page_count (:obj:`int`, optional):
                Only fetch up to this many pages. If 0, get all pages.

                Defaults to: 0.
            max_row_count (:obj:`int`, optional):
                Only fetch up to this many rows.

                Defaults to: 0.
            cache_expiration (:obj:`int`, optional):
                Have the API keep the cache_id that is created on initial get
                answers page alive for N seconds.

                Defaults to: 900.
            hashes (:obj:`bool`, optional):
                Have the API include the hashes of rows values

                Defaults to: False.
            sleep (:obj:`int`, optional):
                Wait N seconds between fetching each page.

                Defaults to: 5.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get_result_data`.

        Notes:
            If max_page_count and max_row_count are 0, fetch pages until a page
            returns no answers or the expected row count is hit.

        Returns:
            :obj:`tantrum.api_models.ApiModel`: ResultDataList API Object

        """
        self._check_id()

        start = datetime.datetime.utcnow()

        row_start = 0

        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj
        cmd_args["row_start"] = row_start
        cmd_args["row_count"] = page_size
        cmd_args["cache_expiration"] = cache_expiration
        cmd_args["include_hashes_flag"] = hashes

        result = self.adapter.cmd_get_result_data(**cmd_args)

        self._last_result = result

        datas = result()
        self._last_datas = datas

        data = datas[0]

        cmd_args["cache_id"] = data.cache_id
        cmd_args["row_start"] += page_size

        m = [
            "Received initial answers: {d.rows}",
            "expected row_count: {d.row_count}",
            "estimated total clients: {d.estimated_total}",
        ]
        m = ", ".join(m)
        m = m.format(d=data)
        self.log.info(m)

        all_rows = data.rows
        page_count = 1
        page_rows = all_rows

        while True:

            if len(all_rows or []) >= data.row_count:
                m = "Received expected row_count {c}, considering all answers received"
                m = m.format(c=data.row_count)
                self.log.info(m)
                break

            if not page_rows:
                m = "Received a page with no answers, considering all answers received"
                self.log.info(m)
                break

            if max_page_count and page_count >= max_page_count:
                m = "Reached max page count {c}, considering all answers in"
                m = m.format(c=max_page_count)
                self.log.info(m)
                break

            if max_row_count and len(all_rows or []) >= max_row_count:
                m = "Hit max pages of {max_row_count}, considering all answers received"
                m = m.format(max_row_count=max_row_count)
                self.log.info(m)

            page_count += 1

            page_result = self.adapter.cmd_get_result_data(**cmd_args)
            self._last_result = page_result

            # this should catch errors where API returns result data as None sometimes
            # need to refetch data for N retries if that happens
            page_datas = page_result()
            self._last_datas = page_datas

            page_data = page_datas[0]
            page_rows = page_data.rows

            m = "Received page #{c} answers: {rows}"
            m = m.format(c=page_count, rows=len(page_rows or []))
            self.log.info(m)

            all_rows += page_rows
            cmd_args["row_start"] += page_size

            time.sleep(sleep)

        end = datetime.datetime.utcnow()
        elapsed = end - start

        m = "Finished getting {rows} answers in {dt}"
        m = m.format(rows=len(all_rows or []), dt=elapsed)
        self.log.info(m)

        return datas

    def answers_sse_start_xml(self, hashes=False, **kwargs):
        """Start up a server side export for XML format and get an export_id.

        Args:
            hashes (:obj:`bool`, optional):
                Have the API include the hashes of rows values

                Defaults to: False.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get_result_data`.

        Returns:
            :obj:`str`:

        """
        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj
        cmd_args["export_flag"] = True
        cmd_args["export_format"] = 1
        cmd_args["include_hashes_flag"] = hashes

        result = self.adapter.cmd_get_result_data(**cmd_args)
        self._last_result = result

        m = ["Received Server Side Export start response for XML format", "code={c}"]
        m = ", ".join(m)
        m = m.format(c=result.status_code)
        self.log.debug(m)

        export_id = result.object_obj["export_id"]

        m = ["Started Server Side for XML format", "export_id={e!r}"]
        m = ", ".join(m)
        m = m.format(e=export_id)
        self.log.info(m)

        return export_id

    def answers_sse_start_csv(
        self, flatten=False, headers=True, hashes=False, **kwargs
    ):
        """Start up a server side export for CSV format and get an export_id.

        Args:
            flatten (:obj:`bool`, optional):
                Flatten CSV rows if possible (single line in each cell)

                Defaults to: False.
            headers (:obj:`bool`, optional):
                Include column headers.

                Defaults to: True.
            hashes (:obj:`bool`, optional):
                Have the API include the hashes of rows values

                Defaults to: False.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get_result_data`.

        Returns:
            :obj:`str`:

        """
        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj
        cmd_args["export_flag"] = True
        cmd_args["export_format"] = 3 if flatten else 0
        cmd_args["export_hide_csv_header_flag"] = False if headers else True
        cmd_args["include_hashes_flag"] = hashes

        result = self.adapter.cmd_get_result_data(**cmd_args)
        self._last_result = result

        m = ["Received Server Side Export start response for CSV format", "code={c}"]
        m = ", ".join(m)
        m = m.format(c=result.status_code)
        self.log.debug(m)

        export_id = result.object_obj["export_id"]

        m = ["Started Server Side for CSV format", "export_id={e!r}"]
        m = ", ".join(m)
        m = m.format(e=export_id)
        self.log.info(m)

        return export_id

    def answers_sse_start_cef(self, leading="", trailing="", **kwargs):
        """Start up a server side export for CEF format and get an export_id.

        Args:
            leading (:obj:`str`, optional):
                Prepend this text to each line.

                Defaults to: "".
            trailing (:obj:`str`, optional):
                Append this text to each line.

                Defaults to: "".
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_get_result_data`.

        Returns:
            :obj:`str`:

        """
        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = self.obj
        cmd_args["export_flag"] = True
        cmd_args["export_format"] = 2

        if leading:
            cmd_args["export_leading_text"] = leading

        if trailing:
            cmd_args["export_trailing_text"] = trailing

        result = self.adapter.cmd_get_result_data(**cmd_args)
        self._last_result = result

        m = ["Received Server Side Export start response for CEF format", "code={c}"]
        m = ", ".join(m)
        m = m.format(c=result.status_code)
        self.log.debug(m)

        export_id = result.object_obj["export_id"]

        m = ["Started Server Side for CEF format", "export_id={e!r}"]
        m = ", ".join(m)
        m = m.format(e=export_id)
        self.log.info(m)

        return export_id

    def answers_sse_get_status(self, export_id, **kwargs):
        """Get the status for this questions server side export.

        Args:
            export_id (:obj:`str`):
                An export id returned from :meth:`sse_start`.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapters.ApiClient`.

        Returns:
            :obj:`dict`:

        """
        client_args = {}
        client_args.update(kwargs)
        client_args["method"] = "get"
        client_args["path"] = "export/{export_id}.status".format(export_id=export_id)
        client_args["data"] = ""

        r = self.adapter.api_client(**client_args)

        status_split = [x.strip().lower() for x in r.text.split(".") if x.strip()]
        status = dict(zip(["status", "progress"], status_split))
        status["export_id"] = export_id

        m = [
            "Received SSE status response: path={r.request.url!r}",
            "code={r.status_code}",
            "status={status}",
        ]
        m = ", ".join(m)
        m = m.format(r=r, status=status)
        self.log.debug(m)

        return status

    def answers_sse_poll(self, export_id, poll_sleep=5, max_poll_count=0, **kwargs):
        """Poll a server side export for completion.

        Args:
            export_id (:obj:`str`):
                An export id returned from :meth:`answers_sse_start_xml` or
                :meth:`answers_sse_start_csv` or :meth:`answers_sse_start_cef`.
            poll_sleep (:obj:`int`, optional):
                Check for answers every N seconds.

                Defaults to: 5.
            max_poll_count (:obj:`int`, optional):
                If not 0, only poll N times.

                Defaults to: 0.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`answers_sse_get_status`.

        Returns:
            :obj:`str`:

        """
        self._check_id()

        start = datetime.datetime.utcnow()

        poll_count = 0

        sse_args = {}
        sse_args.update(kwargs)
        sse_args["export_id"] = export_id

        status = self.answers_sse_get_status(**sse_args)

        while True:
            poll_count += 1

            if max_poll_count and poll_count >= max_poll_count:
                m = [
                    "Server Side Export completed",
                    "reached max poll count {c}",
                    "status {status}",
                ]
                m = ", ".join(m)
                m = m.format(c=max_poll_count, status=status)
                self.log.info(m)
                break

            if status["status"] == "completed":
                m = "Server Side Export completed: {status}"
                m = m.format(status=status)
                self.log.info(m)
                break

            if status["status"] == "failed":
                m = "Server Side Export failed: {status}"
                m = m.format(status=status)
                raise exceptions.ModuleError(m)

            time.sleep(poll_sleep)
            status = self.answers_sse_get_status(**sse_args)

        end = datetime.datetime.utcnow()
        elapsed = end - start

        m = "Finished polling for Server Side Export in {dt}, {status}"
        m = m.format(dt=elapsed, status=status)
        self.log.info(m)

        return status

    def answers_sse_get_data(
        self, export_id, return_dict=False, return_obj=True, **kwargs
    ):
        """Get the answers for this question in XML format using server side export.

        Args:
            export_id (:obj:`str`):
                An export id returned from :meth:`sse_start`.
            return_dict (:obj:`bool`, optional):
                If export_id is an XML format, return a dictionary object.

                Defaults to: False.
            return_obj (:obj:`bool`, optional):
                If export_id is XML format, return a ResultSet object.

                Defaults to: True.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapters.ApiClient`.

        Notes:
            If export_id is not XML format or return_dict and return_obj False,
            return the raw text as is.

        Returns:
            :obj:`tantrum.api_models.ApiModel` or :obj:`dict` or :obj:`str`:
                If return_obj = True returns ResultSetList ApiModel object.
                If return_dict = True returns dict.
                Otherwise, return str.

        """
        self._check_id()

        client_args = {}
        client_args.update(kwargs)
        client_args["method"] = "get"
        client_args["path"] = "export/{export_id}.gz".format(export_id=export_id)
        client_args["data"] = ""

        r = self.adapter.api_client(**client_args)

        m = ["Received SSE data response", "code: {r.status_code}", "export_id: {e!r}"]
        m = ", ".join(m)
        m = m.format(r=r, e=export_id)
        self.log.info(m)

        data = r.text

        if "xml" in export_id and (return_dict or return_obj):
            result = results.Soap(
                api_objects=self.api_objects,
                response_body=r.text,
                request_body=r.request.body,
                method=r.request.method,
                url=r.request.url,
                status_code=r.status_code,
                origin=r,
                lvl=self.log.level,
            )
            data = "<{r}>{data}</{r}>".format(data=data, r="result_set")
            src = "SSE get data response"
            data = result.str_to_obj(text=data, src=src, try_int=False)

            if return_dict:
                return data

            data = self.api_objects.ResultSet(**data["result_set"])
            data = self.api_objects.ResultSetList(*[data])
            return data

        return data


class ParsedQuestion(Workflow):
    def __str__(self):
        """Show object info.

        Returns:
            (:obj:`str`)

        """
        ctmpl = "{c.__module__}.{c.__name__}".format
        bits = [
            "parse matches: {c}".format(c=len(self.obj)),
            "has exact match: {em}".format(em=True if self.get_canonical else False),
        ]
        bits = "({})".format(", ".join(bits))
        cls = ctmpl(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    @property
    def get_canonical(self):
        """Return any parse result that is an exact match."""
        for x in self.obj:
            if x.question.from_canonical_text:
                return x
        return None

    def map_select_params(self, pq):
        """Map parameters to sensors on the left hand side of the question."""
        param_cls = self.api_objects.Parameter
        param_values = pq.parameter_values
        selects = pq.question.selects or []

        for select in selects:
            if not param_values:
                m = "No more parameter values left to map"
                self.log.debug(m)
                return

            sensor = select.sensor

            if not sensor.parameter_definition:
                m = "No parameters defined on sensor {s}, going to next"
                m = m.format(s=sensor)
                self.log.debug(m)
                continue

            sensor.source_id = sensor.id
            sensor.id = None
            sensor.parameters = self.api_objects.ParameterList()
            params = json.loads(sensor.parameter_definition)["parameters"]

            for param in params:
                if not param_values:
                    m = "No more parameter values left to map"
                    self.log.debug(m)
                    return

                key = "||{}||".format(param["key"])
                value = param_values.pop(0)
                sensor.parameters.append(param_cls(key=key, value=value))

                m = "Mapped parameter {k!r}='{v}' for {s}"
                m = m.format(k=key, v=value, s=sensor)
                self.log.debug(m)

    def map_group_params(self, pq, group):
        """Map parameters to filters on the right hand side of the question."""
        param_cls = self.api_objects.Parameter
        group_sensors = pq.question_group_sensors
        param_values = pq.parameter_values

        if not group:
            m = "Empty group, not mapping group params"
            self.log.debug(m)
            return

        if not group_sensors:
            m = "No question group sensors defined, not mapping group params"
            self.log.debug(m)
            return

        for group_filter in group.filters or []:
            if not param_values:
                m = "No more parameter values left to map"
                self.log.debug(m)
                return

            m = "Now mapping parameters for group filter: {gf}"
            m = m.format(gf=group_filter)
            self.log.debug(m)

            sensor_id = group_filter.sensor.id
            sensor = [x for x in group_sensors if x.id == sensor_id][0]

            if not sensor.parameter_definition:
                m = "No parameters defined on sensor {s}, going to next"
                m = m.format(s=sensor)
                self.log.debug(m)
                continue

            sensor.source_id = sensor.id
            sensor.id = None
            sensor.parameters = self.api_objects.ParameterList()
            params = json.loads(sensor.parameter_definition)["parameters"]
            for param in params:
                if not param_values:
                    m = "No more parameter values left to map"
                    self.log.debug(m)
                    return

                key = "||{}||".format(param["key"])
                value = param_values.pop(0)
                sensor.parameters.append(param_cls(key=key, value=value))

                m = "Mapped parameter {k!r}='{v}' for {s}"
                m = m.format(k=key, v=value, s=sensor)
                self.log.debug(m)

            group_filter.sensor = sensor

        for sub_group in group.sub_groups or []:
            self.map_group_params(pq, sub_group)

    @property
    def result_indexes(self):
        """Get the parse result indices in str form."""
        pq_tmpl = "  index: {idx}, result: {text!r}, params: {params}, exact: {exact}"
        pq_tmpl = pq_tmpl.format

        pq_list = []
        for idx, pq in enumerate(self.obj):
            pq_txt = pq_tmpl(
                idx=idx,
                text=pq.question_text,
                params=list(pq.parameter_values or []),
                exact=bool(pq.question.from_canonical_text),
            )
            pq_list.append(pq_txt)
        return "\n".join(pq_list)

    def pick(self, index=None, use_exact_match=True, use_first=False, **kwargs):
        """Pick a parse result and ask it.

        Args:
            index (:obj:`int`, optional):
                Index of parse result to ask.

                Defaults to: None.
            use_exact_match (:obj:`bool`, optional):
                If index is None and one of the parse results is an exact match,
                pick and ask it.

                Defaults to: True.
            use_first (:obj:`bool`, optional):
                If index is None and there is no exact match,
                pick the first parse result and ask it.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_add_parsed_question`.

        Returns:
            :obj:`Question`

        """
        if index:
            pq = self.obj[index]

            m = "Picking parsed question based on index {index}: {pq.question}"
            m = m.format(index=index, pq=pq)
            self.log.info(m)

        elif use_exact_match and self.get_canonical:
            pq = self.get_canonical

            m = "Picking parsed question based on exact match: {pq.question}"
            m = m.format(pq=pq)
            self.log.info(m)

        elif use_first:
            pq = self.obj[0]

            m = "Picking first matching parsed question: {pq.question}"
            m = m.format(pq=pq)
            self.log.info(m)

        else:

            err = [
                "No index supplied",
                "no exact matching parsed result",
                "and use_first is False!",
            ]
            err = ", ".join(err)
            err = [err, "Supply an index of a parsed result:", self.result_indexes]
            err = "\n".join(err)

            raise exceptions.ModuleError(err)

        self.map_select_params(pq=pq)

        m = "Finished mapping parameters for selects, parameter values left: {pv!r}"
        m = m.format(pv=pq.parameter_values)
        self.log.debug(m)

        self.map_group_params(pq=pq, group=pq.question.group)

        m = "Finished mapping parameters for groups, parameter values left: {pv!r}"
        m = m.format(pv=pq.parameter_values)
        self.log.debug(m)

        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["obj"] = pq

        result = self.adapter.cmd_add_parsed_question(**cmd_args)

        result_obj = result()

        workflow = Question(
            adapter=self.adapter, obj=result_obj, lvl=self.log.level, result=result
        )

        m = "Added parsed question: {w}"
        m = m.format(w=workflow)
        self.log.info(m)

        workflow.refetch()
        return workflow

    @classmethod
    def parse(cls, adapter, text, lvl="info", **kwargs):
        """Get parse results of text from API.

        Args:
            adapter (:obj:`tantrum.adapters.Adapter`):
                Adapter to use for this workflow.
            text (:obj:`str`):
                Question text to parse.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`tantrum.adapter.Adapter.cmd_parse_question`.

        Returns:
            :obj:`ParsedQuestion`

        """
        log = utils.logs.get_obj_log(obj=cls, lvl=lvl)

        cmd_args = {}
        cmd_args.update(kwargs)
        cmd_args["text"] = text

        result = adapter.cmd_parse_question(**cmd_args)
        result_obj = result()
        if result_obj is None:
            m = "No parse results returned for text: {t!r}"
            m = m.format(t=text)
            raise exceptions.ModuleError(m)
        any_canonical = any([x.question.from_canonical_text for x in result_obj])

        m = "Received {n} parse results (any exact match: {ac})"
        m = m.format(n=len(result_obj), ac=any_canonical)
        log.info(m)

        return cls(adapter=adapter, obj=result_obj, lvl=lvl, result=result)


OPERATOR_MAP = {
    "less": {"op": "Less", "tmpl": "{value}"},
    "lessequal": {"op": "LessEqual", "tmpl": "{value}"},
    "greater": {"op": "Greater", "tmpl": "{value}"},
    "greaterequal": {"op": "GreaterEqual", "tmpl": "{value}"},
    "equal": {"op": "Equal", "tmpl": "{value}"},
    "regex": {"op": "RegexMatch", "tmpl": "{value}"},
    "startswith": {"op": "RegexMatch", "tmpl": ".*{value}"},
    "endswith": {"op": "RegexMatch", "tmpl": "{value}.*"},
    "contains": {"op": "RegexMatch", "tmpl": ".*{value}.*"},
    "hash": {"op": "HashMatch", "tmpl": "{value}"},
}


TYPE_MAP = {
    "Hash": 0,
    # SENSOR_RESULT_TYPE_STRING
    "String": 1,
    # SENSOR_RESULT_TYPE_VERSION
    "Version": 2,
    # SENSOR_RESULT_TYPE_NUMERIC
    "NumericDecimal": 3,
    # SENSOR_RESULT_TYPE_DATE_BES
    "BESDate": 4,
    # SENSOR_RESULT_TYPE_IPADDRESS
    "IPAddress": 5,
    # SENSOR_RESULT_TYPE_DATE_WMI
    "WMIDate": 6,
    #  e.g. "2 years, 3 months, 18 days, 4 hours, 22 minutes:
    # 'TimeDiff', and 3.67 seconds" or "4.2 hours"
    # (numeric + "Y|MO|W|D|H|M|S" units)
    "TimeDiff": 7,
    #  e.g. 125MB or 23K or 34.2Gig (numeric + B|K|M|G|T units)
    "DataSize": 8,
    "NumericInteger": 9,
    "VariousDate": 10,
    "RegexMatch": 11,
    "LastOperatorType": 12,
}

PCT_FMT = "{0:.0f}%".format


def get_operator_map(operator):
    """Validate operator against :data:`OPERATOR_MAP`."""
    if operator in OPERATOR_MAP:
        return OPERATOR_MAP[operator]
    m = "Operator {o!r} is invalid, must be one of {vo}"
    m = m.format(o=operator, vo=list(OPERATOR_MAP.keys()))
    raise exceptions.ModuleError(m)


def get_type_map(type):
    """Validate type against :data:`TYPE_MAP`."""
    if type in TYPE_MAP:
        return TYPE_MAP[type]
    m = "Type {o!r} is invalid, must be one of {vo}"
    m = m.format(o=type, vo=list(TYPE_MAP.keys()))
    raise exceptions.ModuleError(m)



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.api_clients`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


class ModuleError(exceptions.PackageError):
    """Parent of all exceptions for :mod:`tantrum.api_clients`."""

    pass


class ModuleWarning(exceptions.PackageWarning):
    """Parent of all warnings for :mod:`tantrum.api_clients`."""

    pass


class GetPlatformVersionWarning(ModuleWarning):
    """Thrown when an issue happens while trying to get the platform version."""

    pass



# -*- coding: utf-8 -*-
"""Clients for making API requests to Tanium."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import abc
import six
import warnings

from . import exceptions
from .. import utils

WSDL_PATH = "/libraries/taniumjs/console.wsdl"
""":obj:`str`: URL path to find console.wsdl in :func:`get_wsdl`."""

warnings.simplefilter(action="once", category=exceptions.GetPlatformVersionWarning)
# only warn once about issues getting the platform version


@six.add_metaclass(abc.ABCMeta)
class ApiClient(object):
    """Abstract base class for all ApiClients."""

    @abc.abstractproperty
    def auth_method(self):
        """Get the AuthMethod for this object.

        Returns:
            :obj:`tantrum.auth_methods.AuthMethod`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def http_client(self):
        """Get the HTTP Client for this object.

        Returns:
            :obj:`tantrum.http_client.HttpClient`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def url(self):
        """Get the URL from :attr:`ApiClient.http_client`.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def version(self):
        """Get the platform version from :attr:`ApiClient.config`.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def config(self):
        """Get the deserialized config.json from :attr:`ApiClient.url`.

        Returns:
            :obj:`dict`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def info(self):
        """Get the deserialized info.json from :attr:`ApiClient.url`.

        Returns:
            :obj:`dict`

        """
        raise NotImplementedError  # pragma: no cover


class Soap(ApiClient):
    """Client for making SOAP API requests to Tanium."""

    def __init__(self, http_client, auth_method, lvl="info"):
        """Constructor.

        Args:
            http_client (:obj:`tantrum.http_client.HttpClient`):
                Object for sending HTTP requests.
            auth_method (:obj:`tantrum.auth_methods.AuthMethod`):
                Object for sending login and logout requests.
            lvl (:obj:`str`, optional):
                Logging level for this object.

                Defaults to: "info".

        """
        self.log = utils.logs.get_obj_log(obj=self, lvl=lvl)
        """:obj:`logging.Logger`: Log for this object."""
        self._auth_method = auth_method
        self._http_client = http_client

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = ["url={!r}".format(self.url)]
        bits = "({})".format(", ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    def __call__(
        self,
        data,
        path="/soap",
        method="post",
        timeout=30,
        headers=None,
        verify=None,
        save_last=None,
        save_history=None,
        log_request=None,
        log_response=None,
        cause="",
    ):
        """Get response of POST of data to /soap and return a response object.

        Args:
            data (:obj:`str`):
                Body to send in request.
            timeout (:obj:`int`, optional):
                Response timeout.

                Defaults to: 30.
            headers (:obj:`dict`):
                Headers to send in request.

                Defaults to: None.
            verify (:obj:`bool` or :obj:`str`, optional):
                Enable SSL certification validation.
                If None uses :attr:`tantrum.http_client.HttpClient.verify`.

                Defaults to: None.
            save_last (:obj:`bool`, optional):
                Save last request to :attr:`HttpClient.last_request` and last response
                to :attr:`HttpClient.last_response`.
                If None uses :attr:`tantrum.http_client.HttpClient.save_last`.

                Defaults to: None.
            save_history (:obj:`bool`, optional):
                Append last response to :attr:`HttpClient.history`.
                If None uses :attr:`tantrum.http_client.HttpClient.save_history`.

                Defaults to: None.
            log_request (:obj:`bool`, optional):
                Log request details to debug level.
                If None uses :attr:`tantrum.http_client.HttpClient.log_request`.

                Defaults to: None.
            log_response (:obj:`bool`, optional):
                Log response details to debug level.
                If None uses :attr:`tantrum.http_client.HttpClient.log_response`.

                Defaults to: None.
            cause (:obj:`str`, optional):
                String to explain purpose of request.

                Defaults to: "".


        Returns:
            :obj:`requests.Response`

        """
        headers = headers or {}
        headers = {k: v for k, v in headers.items()}
        headers.update(self.auth_method.token_headers)
        r = self.http_client(
            method=method,
            path=path,
            data=data,
            headers=headers,
            timeout=timeout,
            verify=verify,
            save_last=save_last,
            save_history=save_history,
            log_request=log_request,
            log_response=log_response,
            cause=cause,
        )
        return r

    @property
    def auth_method(self):
        """Get the AuthMethod for this object.

        Returns:
            :obj:`tantrum.auth_methods.AuthMethod`

        """
        return self._auth_method

    @property
    def http_client(self):
        """Get the HTTP Client for this object.

        Returns:
            :obj:`tantrum.http_client.HttpClient`

        """
        return self._http_client

    @property
    def url(self):
        """Get the URL from :attr:`ApiClient.http_client`.

        Returns:
            :obj:`str`

        """
        return self.http_client.url

    def version(self):
        """Get the platform version from :attr:`ApiClient.config`.

        Returns:
            :obj:`str`

        """
        if not getattr(self, "_version", None):
            self._version = get_version(self.http_client)
        return self._version

    def config(self):
        """Get the deserialized config.json from :attr:`ApiClient.url`.

        Returns:
            :obj:`dict`

        """
        return get_config(self.http_client)

    def info(self):
        """Get the deserialized info.json from :attr:`ApiClient.url`.

        Returns:
            :obj:`dict`

        """
        return self.http_client(
            method="get",
            path="/info.json",
            headers=self.auth_method.token_headers,
            timeout=15,
        ).json()


def get_version(http_client):
    """Get serverVersion key from :func:`get_config`.

    Args:
        http_client (:obj:`tantrum.http_client.HttpClient`):
            Object for sending HTTP request.

    Raises:
        :exc:`exceptions.GetPlatformVersionWarning`:
            On error getting serverVersion key.

    Returns:
        :obj:`str`

    """
    cause = "Get platform version"
    try:
        return get_config(http_client=http_client, cause=cause)["serverVersion"]
    except Exception as exc:
        error = "Failed to get server version from url {url!r}, error: {exc}"
        error = error.format(url=http_client.url, exc=exc)
        warnings.warn(error, exceptions.GetPlatformVersionWarning)
        return ""


def get_wsdl(http_client):
    """Get response of GET to /libraries/taniumjs/console.wsdl.

    Args:
        http_client (:obj:`tantrum.http_client.HttpClient`):
            Object for sending HTTP request.

    Returns:
        :obj:`str`

    """
    path = WSDL_PATH
    return http_client(method="get", path=path, timeout=5).text


def get_config(http_client, cause=""):
    """Get response of GET to /config/console.json.

    Args:
        http_client (:obj:`tantrum.http_client.HttpClient`):
            Object for sending HTTP request.

    Returns:
        :obj:`dict`

    """
    path = "/config/console.json"
    return http_client(method="get", path=path, timeout=5, cause=cause).json()
